from threading import get_ident

from sqlalchemy import engine, orm
from sqlalchemy.orm.exc import UnmappedClassError


class _QueryProperty:
    def __init__(self, sa):
        self.sa = sa

    def __get__(self, obj, type):
        try:
            mapper = orm.class_mapper(type)
            if mapper:
                return type.query_class(mapper, session=self.sa.session())
        except UnmappedClassError:
            return None


class SQLAlchemyDB:
    '''
    Simple SQLAlchemy helper inspired by Flask-SQLAlchemy code.

    Allows the code to use a session bind to Flask context.
    '''

    def __init__(self, db_connection, base_model, session_options=None, **kwargs):
        self.engine = engine.create_engine(db_connection, **kwargs)
        self.Query = orm.Query
        self.session_options = session_options
        self.Model = self.extend_base_model(base_model)
        self._session = None

    def extend_base_model(self, base):
        if not getattr(base, 'query_class', None):
            base.query_class = self.Query

        base.query = _QueryProperty(self)
        return base

    @property
    def session(self):
        if self._session is None:
            self._session = self.create_scoped_session(self.session_options)
        return self._session

    @property
    def metadata(self):
        return self.Model.metadata

    def create_scoped_session(self, options=None):

        options = options or {}

        options.setdefault('query_cls', self.Query)

        return orm.scoped_session(
            self.create_session(options), scopefunc=get_ident)

    def create_session(self, options):
        kwargs = {
            "bind": self.engine,
            **options,
        }
        return orm.sessionmaker(**kwargs)

    def create_all_tables(self):
        self.metadata.create_all(bind=self.engine)

    def init_app(self, app):
        @app.teardown_appcontext
        def shutdown_session(response_or_exc):
            self.session.remove()
            return response_or_exc



"""Database models for taxii2 entities."""
import datetime
import uuid

import sqlalchemy
from opentaxii.persistence.sqldb.common import GUID, UTCDateTime
from opentaxii.taxii2 import entities
from sqlalchemy import literal
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class ApiRoot(Base):
    """Database equivalent of `entities.ApiRoot`."""

    __tablename__ = "opentaxii_api_root"

    id = sqlalchemy.Column(GUID, primary_key=True, default=uuid.uuid4)
    default = sqlalchemy.Column(sqlalchemy.Boolean, nullable=False)
    title = sqlalchemy.Column(sqlalchemy.String(100), nullable=False, index=True)
    description = sqlalchemy.Column(sqlalchemy.Text)
    is_public = sqlalchemy.Column(sqlalchemy.Boolean, nullable=False)

    collections = relationship("Collection", back_populates="api_root")

    def set_default(self, session: sqlalchemy.orm.Session):
        """Set this api root as default. Make sure there is always at most 1 default api root."""
        session.query(ApiRoot).filter(ApiRoot.default == literal(True)).update(
            {ApiRoot.default: literal(False)}
        )
        session.query(ApiRoot).filter(ApiRoot.id == self.id).update(
            {ApiRoot.default: literal(True)}
        )

    @classmethod
    def from_entity(cls, entity: entities.ApiRoot):
        """Generate database model from input entity."""
        return cls(**entity.to_dict())


class Job(Base):
    """Database equivalent of `entities.Job`."""

    __tablename__ = "opentaxii_job"

    id = sqlalchemy.Column(GUID, primary_key=True, default=uuid.uuid4)
    api_root_id = sqlalchemy.Column(
        GUID, sqlalchemy.ForeignKey("opentaxii_api_root.id", ondelete="CASCADE")
    )
    status = sqlalchemy.Column(
        sqlalchemy.Enum("pending", "complete", name="job_status_enum")
    )
    request_timestamp = sqlalchemy.Column(UTCDateTime, nullable=True)
    completed_timestamp = sqlalchemy.Column(UTCDateTime, nullable=True)

    details = relationship("JobDetail", back_populates="job")

    __table_args__ = (
        sqlalchemy.Index("ix_opentaxii_job_api_root_id_id", api_root_id, id),
    )

    @classmethod
    def cleanup(cls, session: sqlalchemy.orm.Session) -> int:
        """
        Remove jobs that are >24h old.

        :return: The number of removed jobs.
        """
        jobs = session.query(cls).filter(
            cls.completed_timestamp
            < (
                datetime.datetime.now(datetime.timezone.utc)
                - datetime.timedelta(hours=24)
            )
        )
        return jobs.delete()

    @classmethod
    def from_entity(cls, entity: entities.Job):
        """Generate database model from input entity."""
        return cls(**entity.to_dict())


class JobDetail(Base):
    """Database equivalent of `entities.JobDetail`."""

    __tablename__ = "opentaxii_job_detail"

    id = sqlalchemy.Column(GUID, primary_key=True, default=uuid.uuid4)
    job_id = sqlalchemy.Column(
        GUID, sqlalchemy.ForeignKey("opentaxii_job.id", ondelete="CASCADE"), index=True
    )
    stix_id = sqlalchemy.Column(sqlalchemy.String(100), index=True)
    version = sqlalchemy.Column(UTCDateTime)
    message = sqlalchemy.Column(sqlalchemy.Text)
    status = sqlalchemy.Column(
        sqlalchemy.Enum("success", "failure", "pending", name="job_detail_status_enum")
    )

    job = relationship("Job", back_populates="details")

    @classmethod
    def from_entity(cls, entity: entities.JobDetail):
        """Generate database model from input entity."""
        return cls(**entity.to_dict())


class Collection(Base):
    """Database equivalent of `entities.Collection`."""

    __tablename__ = "opentaxii_collection"

    id = sqlalchemy.Column(GUID, primary_key=True, default=uuid.uuid4)
    api_root_id = sqlalchemy.Column(
        GUID, sqlalchemy.ForeignKey("opentaxii_api_root.id", ondelete="CASCADE")
    )
    title = sqlalchemy.Column(sqlalchemy.String(100), nullable=False, index=True)
    description = sqlalchemy.Column(sqlalchemy.Text)
    alias = sqlalchemy.Column(sqlalchemy.String(100), nullable=True)
    is_public = sqlalchemy.Column(sqlalchemy.Boolean, nullable=False)

    api_root = relationship("ApiRoot", back_populates="collections")
    objects = relationship("STIXObject", back_populates="collection")

    __table_args__ = (sqlalchemy.UniqueConstraint(api_root_id, alias),)

    @classmethod
    def from_entity(cls, entity: entities.Collection):
        """Generate database model from input entity."""
        return cls(**entity.to_dict())


class STIXObject(Base):
    """Database equivalent of `entities.STIXObject`."""

    __tablename__ = "opentaxii_stixobject"

    pk = sqlalchemy.Column(GUID, primary_key=True, default=uuid.uuid4)
    id = sqlalchemy.Column(sqlalchemy.String(100), index=True)
    collection_id = sqlalchemy.Column(
        GUID,
        sqlalchemy.ForeignKey("opentaxii_collection.id", ondelete="CASCADE"),
        index=True,
    )
    type = sqlalchemy.Column(sqlalchemy.String(50), index=True)
    spec_version = sqlalchemy.Column(sqlalchemy.String(10), index=True)  # STIX version
    date_added = sqlalchemy.Column(UTCDateTime, index=True)
    version = sqlalchemy.Column(UTCDateTime, index=True)
    serialized_data = sqlalchemy.Column(sqlalchemy.JSON)

    collection = relationship("Collection", back_populates="objects")

    __table_args__ = (
        sqlalchemy.UniqueConstraint(collection_id, id, version),
        sqlalchemy.Index("ix_opentaxii_stixobject_date_added_id", date_added, id),
    )

    @classmethod
    def from_entity(cls, entity: entities.STIXObject):
        """Generate database model from input entity."""
        return cls(**entity.to_dict())



import base64
import binascii
import functools
import importlib
import inspect
import logging
import re
import sys
from typing import Optional, Tuple

import structlog
from six.moves import urllib

from .entities import Account
from .exceptions import InvalidAuthHeader
from .taxii.converters import dict_to_service_entity
from .taxii.entities import CollectionEntity, deserialize_content_bindings

log = structlog.getLogger(__name__)


def get_path_and_address(domain, address):
    parsed = urllib.parse.urlparse(address)
    if parsed.scheme:
        return None, address
    return address, domain + address


def import_class(module_class_name):
    module_name, _, class_name = module_class_name.rpartition(".")
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def initialize_api(api_config):
    class_name = api_config["class"]
    cls = import_class(class_name)
    params = api_config.get("parameters", None)

    if params and inspect.signature(cls).parameters:
        instance = cls(**params)
    else:
        instance = cls()

    log.info("api.initialized", api=class_name)

    return instance


def parse_basic_auth_token(token):
    try:
        value = base64.b64decode(token)
    except (TypeError, binascii.Error):
        raise InvalidAuthHeader("Can't decode Basic Auth header value")

    try:
        value = value.decode("utf-8")
        username, password = value.split(":", 1)
        return (username, password)
    except ValueError:
        raise InvalidAuthHeader("Invalid Basic Auth header value")


class PlainRenderer:
    def __call__(self, logger, name, event_dict):
        details = event_dict.copy()
        timestamp = details.pop("timestamp")
        logger = details.pop("logger")
        level = details.pop("level")
        event = details.pop("event")
        pairs = ", ".join(["%s=%s" % (k, v) for k, v in details.items()])
        return "{timestamp} [{logger}] {level}: {event} {pairs}".format(
            timestamp=timestamp,
            logger=logger,
            level=level,
            event=event,
            pairs=("{{{}}}".format(pairs) if pairs else ""),
        )


def configure_logging(logging_levels, plain=False, stream=sys.stderr):
    renderer = PlainRenderer() if plain else structlog.processors.JSONRenderer()

    attr_processors = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.TimeStamper(fmt="iso"),
    ]

    structlog.configure_once(
        processors=(
            [structlog.stdlib.filter_by_level]
            + attr_processors
            + [
                structlog.stdlib.PositionalArgumentsFormatter(),
                structlog.processors.StackInfoRenderer(),
                structlog.processors.format_exc_info,
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ]
        ),
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    formatter = structlog.stdlib.ProcessorFormatter(
        processor=renderer, foreign_pre_chain=attr_processors
    )

    handler = AtomicStreamHandler(stream)
    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.handlers = [handler]

    # warnings issued by the ``warnings`` module will be
    # redirected to the ``py.warnings`` logger
    logging.captureWarnings(True)

    for logger, level in logging_levels.items():

        if logger.lower() == "root":
            logger = ""

        logging.getLogger(logger).setLevel(level.upper())


class AtomicStreamHandler(logging.StreamHandler):
    """A ``StreamHandler`` override that provides ``emit`` method which
    does only one ``write`` per record to the underlying stream. This
    is to keep an individual logging record per line. It's important
    for JSON logging.

    In our case we have pre-forked gunicorn worker processes which
    inherit ``stdout`` from the master process and write into it
    concurrently [1]_. Kernel has the following guarantee for
    write [2]_::

        The adjustment of the file offset and the write operation
        are performed as an atomic step.

    On Python side, ``_io.TextIOWrapper``, which is the type of
    ``sys.stdout``, doesn't provide explicit guarantee but experiments
    show that for records of 8KB and 50 forked processes, lines are
    not interleaved.

    Output stream buffering [3]_ doesn't seem to have an effect.

    .. [1]: https://github.com/benoitc/gunicorn/issues/1272
    .. [2]: https://linux.die.net/man/2/write
    .. [3]: https://eklitzke.org/stdout-buffering
    """

    def emit(self, record):
        try:
            msg = self.format(record)
            self.stream.write("%s\n" % msg)
            self.flush()
        except Exception:
            self.handleError(record)


def sync_conf_dict_into_db(server, config, force_collection_deletion=False):
    services = config.get("services", [])
    sync_services(server.servers.taxii1, services)
    collections = config.get("collections", [])
    sync_collections(
        server.servers.taxii1, collections, force_deletion=force_collection_deletion
    )
    accounts = config.get("accounts", [])
    sync_accounts(server, accounts)


def sync_services(server, services):
    manager = server.persistence

    defined_by_id = {s["id"]: s for s in services}
    existing_by_id = {s.id: s for s in manager.get_services()}

    created_counter = 0
    updated_counter = 0

    for service in services:
        existing = existing_by_id.get(service["id"])
        if existing:
            properties = service.copy()
            properties.pop("id")
            existing.type = properties.pop("type")
            existing.properties = properties
            existing = manager.update_service(existing)
            log.info("sync_services.updated", id=existing.id)
            updated_counter += 1
        else:
            service = dict_to_service_entity(service)
            sobj = manager.create_service(service)
            log.info("sync_services.created", id=sobj.id)
            created_counter += 1

    deleted_counter = 0
    missing_ids = set(existing_by_id.keys()) - set(defined_by_id.keys())
    for missing_id in missing_ids:
        manager.delete_service(missing_id)
        deleted_counter += 1
        log.info("sync_services.deleted", id=missing_id)

    log.info(
        "sync_services.stats",
        updated=updated_counter,
        created=created_counter,
        deleted=deleted_counter,
    )


def sync_collections(server, collections, force_deletion=False):
    manager = server.persistence

    defined_by_name = {c["name"]: c for c in collections}
    existing_by_name = {c.name: c for c in manager.get_collections()}

    created_counter = 0
    updated_counter = 0

    for collection in collections:
        existing = existing_by_name.get(collection["name"])
        collection_data = collection.copy()
        service_ids = collection_data.pop("service_ids")
        if existing:
            collection_data.pop("id", None)
            bindings = deserialize_content_bindings(
                collection_data.pop("supported_content", [])
            )
            for k, v in collection_data.items():
                setattr(existing, k, v)
            existing.supported_content = bindings
            cobj = manager.update_collection(existing)
            manager.set_collection_services(cobj.id, service_ids)
            log.info("sync_collections.updated", name=cobj.name, id=cobj.id)
            updated_counter += 1
        else:
            cobj = manager.create_collection(CollectionEntity(**collection_data))
            manager.set_collection_services(cobj.id, service_ids)
            log.info("sync_collections.created", name=cobj.name, id=cobj.id)
            created_counter += 1

    disabled_counter = 0
    deleted_counter = 0
    missing_names = set(existing_by_name.keys()) - set(defined_by_name.keys())
    for name in missing_names:
        if force_deletion:
            manager.delete_collection(name)
            deleted_counter += 1
            log.info("sync_collections.deleted", name=name)
        else:
            collection = existing_by_name[name]
            collection.available = False
            manager.update_collection(collection)
            disabled_counter += 1
            log.info("sync_collections.disabled", name=name)
    log.info(
        "sync_collections.stats",
        updated=updated_counter,
        created=created_counter,
        disabled=disabled_counter,
        deleted=deleted_counter,
    )


def sync_accounts(server, accounts):
    manager = server.auth

    defined_by_username = {a["username"]: a for a in accounts}
    existing_by_username = {a.username: a for a in manager.get_accounts()}

    created_counter = 0
    updated_counter = 0
    for account in accounts:
        existing = existing_by_username.get(account["username"])
        if existing:
            properties = account.copy()
            password = properties.pop("password")
            existing.permissions = properties.get("permissions", {})
            existing.is_admin = properties.get("is_admin", False)
            existing = manager.update_account(existing, password)
            log.info("sync_accounts.updated", username=existing.username)
            updated_counter += 1
        else:
            obj = Account(
                id=None,
                username=account["username"],
                permissions=account.get("permissions", {}),
                is_admin=account.get("is_admin", False),
            )
            obj = manager.update_account(obj, account["password"])
            log.info("sync_accounts.created", username=obj.username)
            created_counter += 1
    deleted_counter = 0
    missing_usernames = set(existing_by_username.keys()) - set(
        defined_by_username.keys()
    )
    for username in missing_usernames:
        manager.delete_account(username)
        deleted_counter += 1
        log.info("sync_accounts.deleted", username=username)

    log.info(
        "sync_accounts.stats",
        updated=updated_counter,
        created=created_counter,
        deleted=deleted_counter,
    )


def register_handler(
    url_re: str,
    valid_methods: Optional[Tuple[str]] = None,
    valid_accept_mimetypes: Optional[Tuple[str]] = None,
    valid_content_types: Optional[Tuple[str]] = None,
    handles_own_auth: bool = False,
):
    """
    Register decorated method as handler function for `url_re`.

    :param str url_re: The regex to trigger the handler on
    :param list valid_methods: The list of methods to accept for this handler, defaults to ("GET",)
    :param list valid_accept_mimetypes:
        The list of accepted mimetypes to accept for this handler, defaults to
        ("application/taxii+json;version=2.1",)
    :param list valid_content_types:
        The list of content types to accept for this handler, defaults to
        ("application/json",)
    """
    if valid_methods is None:
        valid_methods = ("GET",)
    if valid_accept_mimetypes is None:
        valid_accept_mimetypes = ("application/taxii+json;version=2.1",)
    if valid_content_types is None:
        valid_content_types = ("application/json",)

    def inner_decorator(method):
        @functools.wraps(method)
        def inner(*args, **kwargs):
            return method(*args, **kwargs)

        inner.registered_url_re = re.compile(url_re)
        inner.registered_valid_methods = valid_methods
        inner.registered_valid_accept_mimetypes = valid_accept_mimetypes
        inner.registered_valid_content_types = valid_content_types
        inner.handles_own_auth = handles_own_auth
        return inner

    return inner_decorator



from datetime import datetime

import pytz
import structlog
from lxml import etree
from libtaxii.common import set_xml_parser

from .exceptions import BadMessageStatus
from .bindings import MESSAGE_VALIDATOR_PARSER

log = structlog.getLogger(__name__)


def get_utc_now():
    return datetime.utcnow().replace(tzinfo=pytz.UTC)


def is_content_supported(supported_bindings, content_binding, version=None):

    if not hasattr(content_binding, 'binding_id') or version == 10:
        binding_id = content_binding
        subtype = None
    else:
        binding_id = content_binding.binding_id

        # FIXME: may be not the best option
        subtype = (
            content_binding.subtype_ids[0] if content_binding.subtype_ids
            else None)

    matches = [
        ((supported.binding == binding_id) and
         (not supported.subtypes or subtype in supported.subtypes))
        for supported in supported_bindings
    ]

    return any(matches)


def parse_message(content_type, body, do_validate=True):

    validator_parser = MESSAGE_VALIDATOR_PARSER[content_type]

    if do_validate:
        try:
            result = validator_parser.validator.validate_string(body)
            if not result.valid:
                errors = '; '.join([str(err) for err in result.error_log])
                raise BadMessageStatus(
                    'Request was not schema valid: "{}" for content type "{}"'
                    .format(errors, content_type))
        except etree.XMLSyntaxError as e:
            log.error("Invalid XML received", exc_info=True)
            raise BadMessageStatus('Request was invalid XML', e=e)

    taxii_message = validator_parser.parser(body)

    return taxii_message


def configure_libtaxii_xml_parser(huge_tree=False):
    '''
    Set custom XML parser as a default libtaxii parser
    '''
    # set XML parser in libraxii right away
    set_xml_parser(etree.XMLParser(
        # inject default attributes from DTD or XMLSchema
        attribute_defaults=False,
        # validate against a DTD referenced by the document
        dtd_validation=False,
        # use DTD for parsing
        load_dtd=False,
        # prevent network access for related files (default: True)
        no_network=True,
        # clean up redundant namespace declarations
        ns_clean=True,
        # try hard to parse through broken XML
        recover=False,
        # discard blank text nodes that appear ignorable
        remove_blank_text=False,
        # discard comments
        remove_comments=False,
        #  discard processing instructions
        remove_pis=False,
        # replace CDATA sections by normal text content (default: True)
        strip_cdata=True,
        # save memory for short text content (default: True)
        compact=True,
        # use a hash table of XML IDs for fast access
        # (default: True, always True with DTD validation)
        collect_ids=True,
        # replace entities by their text value (default: True)
        resolve_entities=False,
        # enable/disable security restrictions and support very deep
        # trees and very long text content
        huge_tree=huge_tree))



import base64
import datetime
import json
import uuid
from functools import reduce
from typing import Dict, List, Optional, Tuple

import six
import structlog
from opentaxii.common.sqldb import BaseSQLDatabaseAPI
from opentaxii.persistence import (OpenTAXII2PersistenceAPI,
                                   OpenTAXIIPersistenceAPI)
from opentaxii.persistence.sqldb import taxii2models
from opentaxii.taxii2 import entities
from opentaxii.taxii2.utils import DATETIMEFORMAT
from sqlalchemy import and_, func, literal, or_
from sqlalchemy.orm import Query, load_only

from . import converters as conv
from .models import (Base, ContentBlock, DataCollection, InboxMessage,
                     ResultSet, Service, Subscription)

__all__ = ["SQLDatabaseAPI"]

log = structlog.getLogger(__name__)

YIELD_PER_SIZE = 100


class SQLDatabaseAPI(BaseSQLDatabaseAPI, OpenTAXIIPersistenceAPI):
    """SQL database implementation of OpenTAXII Persistence API.

    Implementation will work with any DB supported by SQLAlchemy package.

    Note: this implementation ignores ``context.account`` and does not have
    any access rules.

    :param str db_connection: a string that indicates database dialect and
                          connection arguments that will be passed directly
                          to :func:`~sqlalchemy.engine.create_engine` method.

    :param bool create_tables=False: if True, tables will be created in the DB.

    :param engine_parameters=None: if defined, these arguments would be passed to sqlalchemy.create_engine
    """

    BASEMODEL = Base

    def get_services(self, collection_id=None):
        if collection_id:
            collection = self.db.session.query(DataCollection).get(collection_id)
            services = collection.services
        else:
            services = self.db.session.query(Service).all()
        return [conv.to_service_entity(s) for s in services]

    def get_service(self, service_id):
        return conv.to_service_entity(self.db.session.query(Service).get(service_id))

    def update_service(self, obj):
        service = self.db.session.query(Service).get(obj.id) if obj.id else None
        if service:
            service.type = obj.type
            service.properties = obj.properties
        else:
            service = Service(id=obj.id, type=obj.type, properties=obj.properties)
        self.db.session.add(service)
        self.db.session.commit()
        return conv.to_service_entity(service)

    def create_service(self, entity):
        return self.update_service(entity)

    def get_collections(self, service_id=None):
        if service_id:
            service = self.db.session.query(Service).get(service_id)
            collections = service.collections
        else:
            collections = self.db.session.query(DataCollection).all()
        return [conv.to_collection_entity(c) for c in collections]

    def get_collection(self, name, service_id=None):
        if service_id:
            collection = (
                self.db.session.query(DataCollection)
                .join(Service.collections)
                .filter(Service.id == service_id)
                .filter(DataCollection.name == name)
                .one_or_none()
            )
        else:
            collection = (
                self.db.session.query(DataCollection)
                .filter(DataCollection.name == name)
                .one_or_none()
            )
        if collection:
            return conv.to_collection_entity(collection)

    def update_collection(self, entity):
        _bindings = conv.serialize_content_bindings(entity.supported_content)
        collection = self.db.session.query(DataCollection).get(entity.id)
        if not collection:
            raise ValueError("DataCollection with id {} is not found".format(entity.id))
        collection.name = entity.name
        collection.type = entity.type
        collection.description = entity.description
        collection.available = entity.available
        collection.accept_all_content = entity.accept_all_content
        collection.bindings = _bindings
        self.db.session.commit()
        return conv.to_collection_entity(collection)

    def delete_collection(self, collection_name):
        collection = (
            self.db.session.query(DataCollection)
            .filter(DataCollection.name == collection_name)
            .one()
        )
        self.db.session.delete(collection)
        self.db.session.commit()

    def delete_service(self, service_id):
        service = self.db.session.query(Service).get(service_id)
        self.db.session.delete(service)
        self.db.session.commit()

    def _get_content_query(
        self,
        collection_id=None,
        start_time=None,
        end_time=None,
        bindings=None,
        count=False,
    ):
        if count:
            query = self.db.session.query(func.count(ContentBlock.id))
        else:
            query = self.db.session.query(ContentBlock).order_by(
                ContentBlock.timestamp_label.asc()
            )

        if collection_id:
            query = query.join(ContentBlock.collections).filter(
                DataCollection.id == collection_id
            )

        if start_time:
            query = query.filter(ContentBlock.timestamp_label > start_time)

        if end_time:
            query = query.filter(ContentBlock.timestamp_label <= end_time)

        if bindings:
            criteria = []
            for binding in bindings:
                if binding.subtypes:
                    criterion = and_(
                        ContentBlock.binding_id == binding.binding,
                        ContentBlock.binding_subtype.in_(binding.subtypes),
                    )
                else:
                    criterion = ContentBlock.binding_id == binding.binding
                criteria.append(criterion)

            query = query.filter(or_(*criteria))

        return query

    def get_content_blocks_count(
        self, collection_id=None, start_time=None, end_time=None, bindings=None
    ):

        query = self._get_content_query(
            collection_id=collection_id,
            start_time=start_time,
            end_time=end_time,
            bindings=bindings,
            count=True,
        )

        return query.scalar()

    def get_content_blocks(
        self,
        collection_id=None,
        start_time=None,
        end_time=None,
        bindings=None,
        offset=0,
        limit=None,
    ):

        query = self._get_content_query(
            collection_id=collection_id,
            start_time=start_time,
            end_time=end_time,
            bindings=bindings,
        )

        query = query.offset(offset)
        if limit:
            query = query.limit(limit)

        return [
            conv.to_block_entity(block) for block in query.yield_per(YIELD_PER_SIZE)
        ]

    def create_collection(self, entity):

        _bindings = conv.serialize_content_bindings(entity.supported_content)

        collection = DataCollection(
            name=entity.name,
            type=entity.type,
            description=entity.description,
            available=entity.available,
            accept_all_content=entity.accept_all_content,
            bindings=_bindings,
        )

        self.db.session.add(collection)
        self.db.session.commit()

        return conv.to_collection_entity(collection)

    def set_collection_services(self, collection_id, service_ids):
        collection = self.db.session.query(DataCollection).get(collection_id)
        if not collection:
            raise ValueError(
                "Collection with id {} does not exist".format(collection_id)
            )
        services = (
            self.db.session.query(Service).filter(Service.id.in_(service_ids)).all()
            if service_ids
            else []
        )
        missing_services = set(service_ids) - {s.id for s in services}
        if missing_services:
            raise ValueError(
                "Services with ids {} do not exist".format(missing_services)
            )
        collection.services = services
        self.db.session.add(collection)
        self.db.session.commit()
        log.debug(
            "collection.services_set",
            id=collection.id,
            name=collection.name,
            services=service_ids,
        )

    def create_inbox_message(self, entity):

        if entity.destination_collections:
            names = json.dumps(entity.destination_collections)
        else:
            names = None

        begin = entity.exclusive_begin_timestamp_label
        end = entity.inclusive_end_timestamp_label

        content = (
            entity.original_message.encode("utf-8")
            if isinstance(entity.original_message, six.string_types)
            else entity.original_message
        )

        message = InboxMessage(
            message_id=entity.message_id,
            original_message=content,
            content_block_count=entity.content_block_count,
            destination_collections=names,
            service_id=entity.service_id,
            result_id=entity.result_id,
            record_count=entity.record_count,
            partial_count=entity.partial_count,
            subscription_collection_name=entity.subscription_collection_name,
            subscription_id=entity.subscription_id,
            exclusive_begin_timestamp_label=begin,
            inclusive_end_timestamp_label=end,
        )

        self.db.session.add(message)
        self.db.session.commit()

        return conv.to_inbox_message_entity(message)

    def create_content_block(self, entity, collection_ids=None, service_id=None):

        if entity.content_binding:
            binding = entity.content_binding.binding
            subtype = (
                entity.content_binding.subtypes[0]
                if entity.content_binding.subtypes
                else None
            )
        else:
            binding = None
            subtype = None

        content = (
            entity.content.encode("utf-8")
            if isinstance(entity.content, six.string_types)
            else entity.content
        )

        content = ContentBlock(
            timestamp_label=entity.timestamp_label,
            inbox_message_id=entity.inbox_message_id,
            content=content,
            binding_id=binding,
            binding_subtype=subtype,
        )

        self.db.session.add(content)
        self.db.session.commit()

        if collection_ids:
            self._attach_content_to_collections(content, collection_ids)

        return conv.to_block_entity(content)

    def _attach_content_to_collections(self, content_block, collection_ids):

        if not collection_ids:
            return

        criteria = DataCollection.id.in_(collection_ids)
        new_collections = self.db.session.query(DataCollection).filter(criteria)

        content_block.collections.extend(new_collections)

        self.db.session.add(content_block)

        log.debug(
            "Content block added to collections",
            content_block=content_block.id,
            collections=new_collections.count(),
        )

        self.db.session.commit()

    def create_result_set(self, entity):

        _bindings = conv.serialize_content_bindings(entity.content_bindings)

        result_set = ResultSet(
            id=entity.id,
            collection_id=entity.collection_id,
            bindings=_bindings,
            begin_time=entity.timeframe[0],
            end_time=entity.timeframe[1],
        )

        self.db.session.add(result_set)
        self.db.session.commit()

        return conv.to_result_set_entity(result_set)

    def get_result_set(self, result_set_id):
        result_set = self.db.session.query(ResultSet).get(result_set_id)
        return conv.to_result_set_entity(result_set)

    def get_subscription(self, subscription_id):
        s = self.db.session.query(Subscription).get(subscription_id)
        return conv.to_subscription_entity(s)

    def get_subscriptions(self, service_id):
        service = self.db.session.query(Service).get(service_id)
        return [conv.to_subscription_entity(s) for s in service.subscriptions]

    def update_subscription(self, entity):

        if entity.params:
            params = dict(
                response_type=entity.params.response_type,
                content_bindings=conv.serialize_content_bindings(
                    entity.params.content_bindings
                ),
            )
        else:
            params = {}

        subscription = (
            self.db.session.query(Subscription).get(entity.subscription_id)
            if entity.subscription_id
            else None
        )

        if subscription:
            subscription.collection_id = entity.collection_id
            subscription.params = json.dumps(params)
            subscription.status = entity.status
            subscription.service_id = entity.service_id
        else:
            subscription = Subscription(
                id=entity.subscription_id,
                collection_id=entity.collection_id,
                params=json.dumps(params),
                status=entity.status,
                service_id=entity.service_id,
            )

        self.db.session.add(subscription)
        self.db.session.commit()

        log.debug(
            "subscription.updated",
            subscription=subscription.id,
            collection=subscription.collection_id,
            status=subscription.status,
        )

        return conv.to_subscription_entity(subscription)

    def create_subscription(self, entity):
        return self.update_subscription(entity)

    def delete_content_blocks(
        self, collection_name, start_time, end_time=None, with_messages=False
    ):

        collection = (
            self.db.session.query(DataCollection)
            .filter_by(name=collection_name)
            .one_or_none()
        )

        if not collection:
            raise ValueError(
                "Collection with name '{}' does not exist".format(collection_name)
            )

        content_blocks_query = (
            self.db.session.query(ContentBlock.id)
            .join(DataCollection.content_blocks)
            .filter(DataCollection.id == collection.id)
            .filter(ContentBlock.timestamp_label > start_time)
        )

        if end_time:
            content_blocks_query = content_blocks_query.filter(
                ContentBlock.timestamp_label <= end_time
            )

        inbox_messages_query = (
            self.db.session.query(InboxMessage.id)
            .join(ContentBlock, ContentBlock.inbox_message_id == InboxMessage.id)
            .filter(ContentBlock.id.in_(content_blocks_query.subquery()))
        )

        if with_messages:
            (
                self.db.session.query(InboxMessage)
                .filter(
                    InboxMessage.id.in_(
                        self.db.session.query(inbox_messages_query.subquery(name="ids"))
                    )
                )
                .delete(synchronize_session=False)
            )

        counter = (
            self.db.session.query(ContentBlock)
            .filter(
                ContentBlock.id.in_(
                    self.db.session.query(content_blocks_query.subquery(name="ids"))
                )
            )
            .delete(synchronize_session=False)
        )

        collection.volume = (
            self.db.session.query(func.count(ContentBlock.id))
            .join(ContentBlock.collections)
            .filter(DataCollection.id == collection.id)
        ).scalar()

        self.db.session.commit()

        return counter


class Taxii2SQLDatabaseAPI(BaseSQLDatabaseAPI, OpenTAXII2PersistenceAPI):
    BASEMODEL = taxii2models.Base

    @staticmethod
    def get_next_param(kwargs: Dict) -> str:
        """
        Get value for `next` based on :class:`Dict` instance.

        :param :class:`Dict` kwargs: The dict to base the `next` param on

        :return: The value to use as `next` param
        :rtype: str
        """
        return base64.b64encode(
            f"{kwargs['date_added'].isoformat()}|{kwargs['id']}".encode("utf-8")
        ).decode()

    @staticmethod
    def parse_next_param(next_param: str) -> Dict:
        """
        Parse provided `next_param` into kwargs to be used to filter stix objects.
        """
        date_added_str, obj_id = (
            base64.b64decode(next_param.encode()).decode().split("|")
        )
        date_added = datetime.datetime.strptime(
            date_added_str.split("+")[0], "%Y-%m-%dT%H:%M:%S.%f"
        ).replace(tzinfo=datetime.timezone.utc)
        return {"id": obj_id, "date_added": date_added}

    def get_api_roots(self) -> List[entities.ApiRoot]:
        query = self.db.session.query(taxii2models.ApiRoot).order_by("title")
        return [
            entities.ApiRoot(
                id=obj.id,
                default=obj.default,
                title=obj.title,
                description=obj.description,
                is_public=obj.is_public,
            )
            for obj in query.all()
        ]

    def get_api_root(self, api_root_id: str) -> Optional[entities.ApiRoot]:
        api_root = (
            self.db.session.query(taxii2models.ApiRoot)
            .filter(taxii2models.ApiRoot.id == api_root_id)
            .one_or_none()
        )
        if api_root:
            return entities.ApiRoot(
                id=api_root.id,
                default=api_root.default,
                title=api_root.title,
                description=api_root.description,
                is_public=api_root.is_public,
            )
        else:
            return None

    def add_api_root(
        self,
        title: str,
        description: Optional[str] = None,
        default: Optional[bool] = False,
        is_public: bool = False,
    ) -> entities.ApiRoot:
        """
        Add a new api root.

        :param str title: Title of the new api root
        :param str description: [Optional] Description of the new api root
        :param bool default: [Optional, False] If the new api should be the default
        :param bool is_public: whether this is a publicly readable API root

        :return: The added ApiRoot entity.
        """
        api_root = taxii2models.ApiRoot(
            title=title, description=description, default=default, is_public=is_public
        )
        self.db.session.add(api_root)
        self.db.session.commit()
        if default:
            api_root.set_default(self.db.session)
        return entities.ApiRoot(
            id=api_root.id,
            default=api_root.default,
            title=api_root.title,
            description=api_root.description,
            is_public=is_public,
        )

    def get_job_and_details(
        self, api_root_id: str, job_id: str
    ) -> Tuple[Optional[entities.Job], List[entities.JobDetail]]:
        job = (
            self.db.session.query(taxii2models.Job)
            .filter(
                taxii2models.Job.api_root_id == api_root_id,
                taxii2models.Job.id == job_id,
            )
            .one_or_none()
        )
        if job is None:
            return None, []
        job_details = (
            self.db.session.query(taxii2models.JobDetail)
            .filter(
                taxii2models.JobDetail.job_id == job_id,
            )
            .order_by(taxii2models.JobDetail.stix_id)
            .all()
        )
        return (
            entities.Job(
                id=job.id,
                api_root_id=job.api_root_id,
                status=job.status,
                request_timestamp=job.request_timestamp,
                completed_timestamp=job.completed_timestamp,
            ),
            [
                entities.JobDetail(
                    id=job_detail.id,
                    job_id=job_detail.job_id,
                    stix_id=job_detail.stix_id,
                    version=job_detail.version,
                    message=job_detail.message,
                    status=job_detail.status,
                )
                for job_detail in job_details
            ],
        )

    def job_cleanup(self) -> int:
        """
        Remove jobs that are >24h old.

        :return: The number of removed jobs.
        """
        return taxii2models.Job.cleanup(self.db.session)

    def get_collections(self, api_root_id: str) -> List[entities.Collection]:
        query = (
            self.db.session.query(taxii2models.Collection)
            .filter(taxii2models.Collection.api_root_id == api_root_id)
            .order_by(taxii2models.Collection.title)
        )
        return [
            entities.Collection(
                id=obj.id,
                api_root_id=obj.api_root_id,
                title=obj.title,
                description=obj.description,
                alias=obj.alias,
                is_public=obj.is_public,
            )
            for obj in query.all()
        ]

    def get_collection(
        self, api_root_id: str, collection_id_or_alias: str
    ) -> Optional[entities.Collection]:
        id_or_alias_filter = taxii2models.Collection.alias == collection_id_or_alias
        try:
            uuid.UUID(collection_id_or_alias)
        except ValueError:
            pass
        else:
            id_or_alias_filter |= taxii2models.Collection.id == collection_id_or_alias
        obj = (
            self.db.session.query(taxii2models.Collection)
            .filter(
                taxii2models.Collection.api_root_id == api_root_id,
                id_or_alias_filter,
            )
            .one_or_none()
        )
        if obj is None:
            return None
        return entities.Collection(
            id=obj.id,
            api_root_id=obj.api_root_id,
            title=obj.title,
            description=obj.description,
            alias=obj.alias,
            is_public=obj.is_public,
        )

    def add_collection(
        self,
        api_root_id: str,
        title: str,
        description: Optional[str] = None,
        alias: Optional[str] = None,
        is_public: bool = False,
    ) -> entities.Collection:
        """
        Add a new collection.

        :param str api_root_id: ID of the api root the new collection is part of
        :param str title: Title of the new collection
        :param str description: [Optional] Description of the new collection
        :param str alias: [Optional] Alias of the new collection
        :param bool is_public: [Optional] Whether collection should be publicly readable

        :return: The added Collection entity.
        """
        collection = taxii2models.Collection(
            api_root_id=api_root_id,
            title=title,
            description=description,
            alias=alias,
            is_public=is_public,
        )
        self.db.session.add(collection)
        self.db.session.commit()

        return entities.Collection(
            id=collection.id,
            api_root_id=collection.api_root_id,
            title=collection.title,
            description=collection.description,
            alias=collection.alias,
            is_public=collection.is_public,
        )

    def _objects_query(self, collection_id: str, ordered: bool) -> Query:
        query = self.db.session.query(taxii2models.STIXObject).filter(
            taxii2models.STIXObject.collection_id == collection_id,
        )
        if ordered:
            query = query.order_by(
                taxii2models.STIXObject.date_added, taxii2models.STIXObject.id
            )
        return query

    def _apply_added_after(
        self, query: Query, added_after: Optional[datetime.datetime] = None
    ) -> Query:
        if added_after is not None:
            query = query.filter(taxii2models.STIXObject.date_added > added_after)
        return query

    def _apply_next_kwargs(
        self, query: Query, next_kwargs: Optional[Dict] = None
    ) -> Query:
        if next_kwargs is not None:
            query = query.filter(
                (taxii2models.STIXObject.date_added > next_kwargs["date_added"])
                | (
                    (taxii2models.STIXObject.date_added == next_kwargs["date_added"])
                    & (taxii2models.STIXObject.id > next_kwargs["id"])
                )
            )
        return query

    def _apply_match_id(
        self, query: Query, match_id: Optional[List[str]] = None
    ) -> Query:
        if match_id is not None:
            query = query.filter(taxii2models.STIXObject.id.in_(match_id))
        return query

    def _apply_match_type(
        self, query: Query, match_type: Optional[List[str]] = None
    ) -> Query:
        if match_type is not None:
            query = query.filter(taxii2models.STIXObject.type.in_(match_type))
        return query

    def _apply_match_version(
        self,
        query: Query,
        collection_id: str,
        match_version: Optional[List[str]] = None,
    ) -> Query:
        if match_version is None:
            match_version = ["last"]
        if "all" in match_version:
            return query
        version_filters = []
        for value in match_version:
            if value == "first":
                min_versions_subq = (
                    self.db.session.query(
                        taxii2models.STIXObject.id,
                        func.min(taxii2models.STIXObject.version).label("min_version"),
                    )
                    .filter(
                        taxii2models.STIXObject.collection_id == collection_id,
                    )
                    .group_by(taxii2models.STIXObject.id)
                    .subquery()
                )
                min_version_pks = (
                    self.db.session.query(taxii2models.STIXObject.pk)
                    .select_from(taxii2models.STIXObject)
                    .join(
                        min_versions_subq,
                        (
                            (taxii2models.STIXObject.id == min_versions_subq.c.id)
                            & (
                                taxii2models.STIXObject.version
                                == min_versions_subq.c.min_version
                            )
                        ),
                    )
                )
                version_filters.append(taxii2models.STIXObject.pk.in_(min_version_pks))
            elif value == "last":
                max_versions_subq = (
                    self.db.session.query(
                        taxii2models.STIXObject.id,
                        func.max(taxii2models.STIXObject.version).label("max_version"),
                    )
                    .filter(
                        taxii2models.STIXObject.collection_id == collection_id,
                    )
                    .group_by(taxii2models.STIXObject.id)
                    .subquery()
                )
                max_version_pks = (
                    self.db.session.query(taxii2models.STIXObject.pk)
                    .select_from(taxii2models.STIXObject)
                    .join(
                        max_versions_subq,
                        (
                            (taxii2models.STIXObject.id == max_versions_subq.c.id)
                            & (
                                taxii2models.STIXObject.version
                                == max_versions_subq.c.max_version
                            )
                        ),
                    )
                )
                version_filters.append(taxii2models.STIXObject.pk.in_(max_version_pks))
            else:
                version_filters.append(taxii2models.STIXObject.version == value)
        query = query.filter(reduce(or_, version_filters))
        return query

    def _apply_match_spec_version(
        self, query: Query, match_spec_version: Optional[List[str]] = None
    ) -> Query:
        if match_spec_version is not None:
            query = query.filter(
                taxii2models.STIXObject.spec_version.in_(match_spec_version)
            )
        return query

    def _apply_limit(
        self, query: Query, limit: Optional[int] = None
    ) -> Tuple[Query, bool]:
        if limit is not None:
            more = limit < query.count()
            query = query.limit(limit)
        else:
            more = False
        return query, more

    def _filtered_objects_query(
        self,
        collection_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_id: Optional[List[str]] = None,
        match_type: Optional[List[str]] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
        ordered: Optional[bool] = True,
    ) -> Tuple[Query, bool]:
        query = self._objects_query(collection_id, ordered)
        query = self._apply_added_after(query, added_after)
        query = self._apply_next_kwargs(query, next_kwargs)
        query = self._apply_match_id(query, match_id)
        query = self._apply_match_type(query, match_type)
        query = self._apply_match_version(query, collection_id, match_version)
        query = self._apply_match_spec_version(query, match_spec_version)
        query, more = self._apply_limit(query, limit)
        return query, more

    def get_manifest(
        self,
        collection_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_id: Optional[List[str]] = None,
        match_type: Optional[List[str]] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[entities.ManifestRecord], bool]:
        query, more = self._filtered_objects_query(
            collection_id=collection_id,
            limit=limit,
            added_after=added_after,
            next_kwargs=next_kwargs,
            match_id=match_id,
            match_type=match_type,
            match_version=match_version,
            match_spec_version=match_spec_version,
        )
        query = query.options(
            load_only(
                taxii2models.STIXObject.id,
                taxii2models.STIXObject.date_added,
                taxii2models.STIXObject.version,
                taxii2models.STIXObject.spec_version,
            )
        )
        return (
            [
                entities.ManifestRecord(
                    id=obj.id,
                    date_added=obj.date_added,
                    version=obj.version,
                    spec_version=obj.spec_version,
                )
                for obj in query.all()
            ],
            more,
        )

    def get_objects(
        self,
        collection_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_id: Optional[List[str]] = None,
        match_type: Optional[List[str]] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[entities.STIXObject], bool, Optional[str]]:
        query, more = self._filtered_objects_query(
            collection_id=collection_id,
            limit=limit,
            added_after=added_after,
            next_kwargs=next_kwargs,
            match_id=match_id,
            match_type=match_type,
            match_version=match_version,
            match_spec_version=match_spec_version,
        )
        items = query.all()
        if more:
            next_param = self.get_next_param(
                {"id": items[-1].id, "date_added": items[-1].date_added}
            )
        else:
            next_param = None
        return (
            [
                entities.STIXObject(
                    id=obj.id,
                    collection_id=collection_id,
                    type=obj.type,
                    spec_version=obj.spec_version,
                    date_added=obj.date_added,
                    version=obj.version,
                    serialized_data=obj.serialized_data,
                )
                for obj in items
            ],
            more,
            next_param,
        )

    def add_objects(
        self, api_root_id: str, collection_id: str, objects: List[Dict]
    ) -> Tuple[entities.Job, List[entities.JobDetail]]:
        job = taxii2models.Job(
            api_root_id=api_root_id,
            status="pending",
            request_timestamp=datetime.datetime.now(datetime.timezone.utc),
        )
        self.db.session.add(job)
        self.db.session.commit()
        job_details = []
        for obj in objects:
            version = datetime.datetime.strptime(
                obj["modified"], DATETIMEFORMAT
            ).replace(tzinfo=datetime.timezone.utc)
            if (
                not self.db.session.query(literal(True))
                .filter(
                    self.db.session.query(taxii2models.STIXObject)
                    .filter(
                        taxii2models.STIXObject.id == obj["id"],
                        taxii2models.STIXObject.collection_id == collection_id,
                        taxii2models.STIXObject.version == version,
                    )
                    .exists()
                )
                .scalar()
            ):
                self.db.session.add(
                    taxii2models.STIXObject(
                        id=obj["id"],
                        collection_id=collection_id,
                        type=obj["id"].split("--")[0],
                        spec_version=obj["spec_version"],
                        date_added=datetime.datetime.now(datetime.timezone.utc),
                        version=version,
                        serialized_data={
                            key: value
                            for (key, value) in obj.items()
                            if key not in ["id", "type", "spec_version"]
                        },
                    )
                )
            job_detail = taxii2models.JobDetail(
                job_id=job.id,
                stix_id=obj["id"],
                version=version,
                message="",
                status="success",
            )
            job_details.append(job_detail)
            self.db.session.add(job_detail)
        job.status = "complete"
        job.completed_timestamp = datetime.datetime.now(datetime.timezone.utc)
        self.db.session.commit()
        return (
            entities.Job(
                id=job.id,
                api_root_id=job.api_root_id,
                status=job.status,
                request_timestamp=job.request_timestamp,
                completed_timestamp=job.completed_timestamp,
            ),
            [
                entities.JobDetail(
                    id=job_detail.id,
                    job_id=job_detail.job_id,
                    stix_id=job_detail.stix_id,
                    version=job_detail.version,
                    message=job_detail.message,
                    status=job_detail.status,
                )
                for job_detail in job_details
            ],
        )

    def get_object(
        self,
        collection_id: str,
        object_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[Optional[List[entities.STIXObject]], bool, Optional[str]]:
        """
        Get single object from database.

        Should return `None` when object matching object_id doesn't exist.
        """
        if (
            not self.db.session.query(literal(True))
            .filter(
                self.db.session.query(taxii2models.STIXObject)
                .filter(
                    taxii2models.STIXObject.id == object_id,
                    taxii2models.STIXObject.collection_id == collection_id,
                )
                .exists()
            )
            .scalar()
        ):
            return (None, False, None)
        query, more = self._filtered_objects_query(
            collection_id=collection_id,
            limit=limit,
            added_after=added_after,
            next_kwargs=next_kwargs,
            match_id=[object_id],
            match_version=match_version,
            match_spec_version=match_spec_version,
        )
        items = query.all()
        if more:
            next_param = self.get_next_param(
                {"id": items[-1].id, "date_added": items[-1].date_added}
            )
        else:
            next_param = None
        return (
            [
                entities.STIXObject(
                    id=obj.id,
                    collection_id=collection_id,
                    type=obj.type,
                    spec_version=obj.spec_version,
                    date_added=obj.date_added,
                    version=obj.version,
                    serialized_data=obj.serialized_data,
                )
                for obj in items
            ],
            more,
            next_param,
        )

    def delete_object(
        self,
        collection_id: str,
        object_id: str,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> None:
        if match_version is None:
            match_version = ["all"]
        query, _ = self._filtered_objects_query(
            collection_id=collection_id,
            match_id=[object_id],
            match_version=match_version,
            match_spec_version=match_spec_version,
            ordered=False,
        )
        query.delete("fetch")

    def get_versions(
        self,
        collection_id: str,
        object_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[entities.VersionRecord], bool]:
        """
        Get all versions of single object from database.

        Should return `None` when object matching object_id doesn't exist.
        """
        if (
            not self.db.session.query(literal(True))
            .filter(
                self.db.session.query(taxii2models.STIXObject)
                .filter(
                    taxii2models.STIXObject.id == object_id,
                    taxii2models.STIXObject.collection_id == collection_id,
                )
                .exists()
            )
            .scalar()
        ):
            return (None, False)
        query, more = self._filtered_objects_query(
            collection_id=collection_id,
            limit=limit,
            added_after=added_after,
            next_kwargs=next_kwargs,
            match_id=[object_id],
            match_version=["all"],
            match_spec_version=match_spec_version,
        )
        query = query.options(
            load_only(
                taxii2models.STIXObject.date_added,
                taxii2models.STIXObject.version,
            )
        )
        return (
            [
                entities.VersionRecord(
                    date_added=obj.date_added,
                    version=obj.version,
                )
                for obj in query.all()
            ],
            more,
        )



import functools
import importlib
import json

try:
    from re import Pattern
except ImportError:
    from typing.re import Pattern

from typing import Callable, ClassVar, NamedTuple, Optional, Tuple, Type

import structlog
from flask import Flask, Response, request
from werkzeug.exceptions import (Forbidden, MethodNotAllowed, NotAcceptable,
                                 NotFound, RequestEntityTooLarge, Unauthorized,
                                 UnsupportedMediaType)

from opentaxii.persistence.exceptions import (DoesNotExistError,
                                              NoReadNoWritePermission,
                                              NoReadPermission,
                                              NoWritePermission)
from opentaxii.taxii2.utils import taxii2_datetimeformat
from opentaxii.taxii2.validation import (validate_delete_filter_params,
                                         validate_envelope,
                                         validate_list_filter_params,
                                         validate_object_filter_params,
                                         validate_versions_filter_params)
from opentaxii.utils import register_handler

from .auth import AuthManager
from .config import ServerConfig
from .entities import Account
from .exceptions import UnauthorizedException
from .local import context
from .persistence import (BasePersistenceManager, Taxii1PersistenceManager,
                          Taxii2PersistenceManager)
from .taxii2.http import make_taxii2_response
from .taxii.bindings import (ALL_PROTOCOL_BINDINGS, MESSAGE_BINDINGS,
                             SERVICE_BINDINGS)
from .taxii.exceptions import (FailureStatus, StatusMessageException,
                               raise_failure)
from .taxii.http import (HTTP_ALLOW, HTTP_X_TAXII_CONTENT_TYPES,
                         get_content_type, get_http_headers,
                         make_taxii_response, validate_request_headers,
                         validate_request_headers_post_parse,
                         validate_response_headers)
from .taxii.services import (CollectionManagementService, DiscoveryService,
                             InboxService, PollService)
from .taxii.services.abstract import TAXIIService
from .taxii.status import process_status_exception
from .taxii.utils import configure_libtaxii_xml_parser, parse_message
from .utils import get_path_and_address, initialize_api

log = structlog.get_logger(__name__)

anonymous_full_access = Account(id=None, username=None, permissions={}, is_admin=True)


class BaseTAXIIServer:
    """
    Base class for common functionality in taxii* servers.
    """

    PERSISTENCE_MANAGER_CLASS: ClassVar[Type[BasePersistenceManager]]
    ENDPOINT_MAPPING: Tuple[(Pattern, Callable[[], Response])]
    app: Flask
    config: dict

    def __init__(self, config: dict):
        self.config = config
        self.persistence = self.PERSISTENCE_MANAGER_CLASS(
            server=self, api=initialize_api(config["persistence_api"])
        )
        self.setup_endpoint_mapping()

    def setup_endpoint_mapping(self):
        mapping = []
        for attr_name in self.__dir__():
            attr = getattr(self, attr_name)
            if hasattr(attr, "registered_url_re"):
                mapping.append((attr.registered_url_re, attr))
        if mapping:
            self.ENDPOINT_MAPPING = tuple(mapping)

    def init_app(self, app: Flask):
        """Connect server and persistence to flask."""
        self.app = app
        self.persistence.api.init_app(app)

    def get_domain(self, service_id):
        """Get domain either from request handler or config."""
        dynamic_domain = self.persistence.get_domain(service_id)
        domain = dynamic_domain or self.config.get("domain")
        return domain

    def get_endpoint(self, relative_path: str) -> Optional[Callable[[], Response]]:
        """Get first endpoint matching relative_path."""
        return

    def handle_internal_error(self, error):
        """
        Handle internal error and return appropriate response.

        Placeholder for subclasses to implement.
        """
        return

    def handle_status_exception(self, error):
        """
        Handle status exception and return appropriate response.

        Placeholder for subclasses to implement.
        """
        return

    def handle_http_exception(self, error):
        return error.get_response()

    def handle_validation_exception(self, error):
        """
        Handle validation exception and return appropriate response.

        Placeholder for subclasses to implement.
        """
        return

    def raise_unauthorized(self):
        """
        Handle unauthorized access.

        Placeholder for subclasses to implement.
        """
        raise UnauthorizedException()


class TAXII1Server(BaseTAXIIServer):
    """TAXII1 Server class.

    This class keeps Presistence API manager instance for TAXII1
    and creates TAXII1 Service instances on request.

    :param dict config:
        OpenTAXII1 server configuration
    """

    TYPE_TO_SERVICE = {
        "inbox": InboxService,
        "discovery": DiscoveryService,
        "collection_management": CollectionManagementService,
        "poll": PollService,
    }
    PERSISTENCE_MANAGER_CLASS = Taxii1PersistenceManager

    def __init__(self, config: dict):
        super().__init__(config)
        signal_hooks = config["hooks"]
        if signal_hooks:
            importlib.import_module(signal_hooks)
            log.info("signal_hooks.imported", hooks=signal_hooks)

        configure_libtaxii_xml_parser(config["xml_parser_supports_huge_tree"])
        log.info("opentaxii.server_configured")

    def _create_services(self, service_entities):
        """Create :class:`TAXIIService` instances from `service_entities`"""

        discovery_services = []
        services = []

        for entity in service_entities:

            _props = dict(entity.properties)
            _props["server"] = self

            _props["path"], _props["address"] = get_path_and_address(
                self.get_domain(entity.id), _props["address"]
            )

            advertised = _props.pop("advertised_services", None)

            if entity.type not in self.TYPE_TO_SERVICE:
                raise ValueError('Unknown service type "%s"' % entity.type)

            service = self.TYPE_TO_SERVICE[entity.type](id=entity.id, **_props)

            services.append(service)

            if advertised:
                discovery_services.append((service, advertised))

        for service, advertised in discovery_services:
            service.set_advertised_services([s for s in services if s.id in advertised])

        return services

    def check_allowed_methods(self):
        valid_methods = ["POST", "OPTIONS"]
        if request.method not in valid_methods:
            raise MethodNotAllowed(valid_methods=valid_methods)

    def get_endpoint(self, relative_path: str) -> Optional[Callable[[], Response]]:
        """Get first endpoint matching relative_path."""
        for endpoint in self.get_services():
            if endpoint.path == relative_path:
                return functools.partial(self.handle_request, endpoint=endpoint)

    def get_services(self, service_ids=None):
        """Get services registered with this TAXII server instance.

        :param list service_ids: list of service IDs (as strings)

        :return: list of services
        :rtype: list of
                :py:class:`opentaxii.taxii.services.abstract.TAXIIService`
        """

        # early return for filtering by empty list of ids
        if service_ids is not None and not service_ids:
            return []

        service_entities = self.persistence.get_services()

        # Services needs to be created all at once to ensure that
        # discovery services list all active advertised services
        services = self._create_services(service_entities)

        if service_ids:
            services = [service for service in services if service.id in service_ids]

        return services

    def get_service(self, id):
        """Get service by ID.

        :param str id: service ID

        :return: service with specified ID or None
        :rtype: :py:class:`opentaxii.taxii.services.abstract.TAXIIService`
        """

        services = self.get_services([id])
        if services:
            return services[0]

    def get_services_for_collection(self, collection, service_type):
        """Get list of services with type ``service_type``, attached
        to collection ``collection``.

        :param `opentaxii.taxii.entities.CollectionEntity` collection:
                    collection in question
        :param str service_type: service type, supported values are
                    listed as keys in :py:attr:`TYPE_TO_SERVICE`

        :return: list of services
        :rtype: list of :py:class:`opentaxii.taxii.services.abstract.TAXIIService`  # noqa
        """

        if service_type not in self.TYPE_TO_SERVICE:
            raise ValueError("Wrong service type: %s" % service_type)

        for_collection = self.persistence.get_services_for_collection(collection)

        ids_for_type = [e.id for e in for_collection if e.type == service_type]

        # Sync services for collection with registered services for this server
        return self.get_services(ids_for_type)

    def handle_request(self, endpoint: TAXIIService):
        """
        Handle request and return appropriate response.

        Process :class:`TAXIIService` with either :meth:`_process_with_service` or :meth:`_process_options_request`.
        """
        self.check_allowed_methods()
        if endpoint.authentication_required and context.account is None:
            raise UnauthorizedException(
                status_type=self.config["unauthorized_status"],
            )

        if not endpoint.authentication_required:
            # if endpoint is not protected, full access
            context.account = anonymous_full_access

        if not endpoint.available:
            raise_failure("The service is not available")

        if request.method == "POST":
            return self._process_with_service(endpoint)
        if request.method == "OPTIONS":
            return self._process_options_request(endpoint)

    @staticmethod
    def _process_options_request(service):
        """
        OPTIONS request, return appropriate allow and content types headers.
        """

        message_bindings = ",".join(service.supported_message_bindings or [])

        return (
            "",
            200,
            {HTTP_ALLOW: "POST, OPTIONS", HTTP_X_TAXII_CONTENT_TYPES: message_bindings},
        )

    @classmethod
    def _process_with_service(cls, service) -> Response:
        """
        POST request, return requested data.
        """

        if "application/xml" not in request.accept_mimetypes:
            raise_failure(
                "The specified values of Accept is not supported: {}".format(
                    ", ".join((request.accept_mimetypes or []))
                )
            )

        validate_request_headers(request.headers, MESSAGE_BINDINGS)

        taxii_message = parse_message(get_content_type(request.headers), request.data)

        try:
            validate_request_headers_post_parse(
                request.headers,
                supported_message_bindings=MESSAGE_BINDINGS,
                service_bindings=SERVICE_BINDINGS,
                protocol_bindings=ALL_PROTOCOL_BINDINGS,
            )
        except StatusMessageException as e:
            e.in_response_to = taxii_message.message_id
            raise e

        response_message = service.process(request.headers, taxii_message)

        response_headers = get_http_headers(response_message.version, request.is_secure)
        validate_response_headers(response_headers)

        taxii_xml = response_message.to_xml(pretty_print=True)
        return make_taxii_response(taxii_xml, response_headers)

    def handle_internal_error(self, error):
        """
        Handle internal error and return appropriate response.
        """
        log.error("Internal error", exc_info=True)

        if "application/xml" not in request.accept_mimetypes:
            return "Unacceptable", 406

        if self.config["return_server_error_details"]:
            message = "Server error occurred: {}".format(error)
        else:
            message = "Server error occurred"

        new_error = FailureStatus(message, e=error)
        xml, headers = process_status_exception(
            new_error, request.headers, request.is_secure
        )
        return make_taxii_response(xml, headers)

    def handle_status_exception(self, error):
        """
        Handle status exception and return appropriate response.
        """
        log.warning("Status exception", exc_info=True)

        if "application/xml" not in request.accept_mimetypes:
            return "Unacceptable", 406

        xml, headers = process_status_exception(
            error, request.headers, request.is_secure
        )
        return make_taxii_response(xml, headers)

    def raise_unauthorized(self):
        """
        Handle unauthorized access.
        """
        raise UnauthorizedException(
            status_type=self.config["unauthorized_status"],
        )


class TAXII2Server(BaseTAXIIServer):
    """
    TAXII2 server class.

    Stub, implementation pending.
    """

    PERSISTENCE_MANAGER_CLASS = Taxii2PersistenceManager

    def handle_http_exception(self, error):
        """Return JSON instead of HTML for HTTP errors."""
        # start with the correct headers and status code from the error
        response = error.get_response()
        # replace the body with JSON
        response.data = json.dumps(
            {
                "code": error.code,
                "name": error.name,
                "description": error.description,
            }
        )
        response.content_type = "application/taxii+json;version=2.1"
        return response

    def handle_validation_exception(self, error):
        """
        Handle validation exception and return appropriate response.
        """
        response = {
            "code": 400,
            "name": "validation error",
            "description": error.messages,
        }
        return make_taxii2_response(response, status=400)

    def get_endpoint(self, relative_path: str) -> Optional[Callable[[], Response]]:
        endpoint = None
        for regex, handler in self.ENDPOINT_MAPPING:
            match = regex.match(relative_path)
            if match:
                endpoint = functools.partial(handler, **match.groupdict())
                break
        if endpoint:
            return functools.partial(self.handle_request, endpoint)

    def check_authentication(self, endpoint: Callable[[], Response]):
        """Check if account is authenticated, unless endpoint handles that itself."""
        if endpoint.func.handles_own_auth:
            # Endpoint will handle auth checks itself
            return
        if context.account is None:
            raise Unauthorized()

    def check_content_length(self):
        if (request.content_length or 0) > self.config["max_content_length"] or len(
            request.data
        ) > self.config[
            "max_content_length"
        ]:  # untestable with flask
            raise RequestEntityTooLarge()

    def check_headers(self, endpoint: Callable[[], Response]):
        if not any(
            [
                valid_accept_mimetype in request.accept_mimetypes
                for valid_accept_mimetype in endpoint.func.registered_valid_accept_mimetypes
            ]
        ):
            raise NotAcceptable()
        if (
            request.method == "POST"
            and request.content_type not in endpoint.func.registered_valid_content_types
        ):
            raise UnsupportedMediaType()

    def check_allowed_methods(self, endpoint: Callable[[], Response]):
        if request.method not in endpoint.func.registered_valid_methods:
            raise MethodNotAllowed(valid_methods=endpoint.func.registered_valid_methods)

    def handle_request(self, endpoint: Callable[[], Response]):
        self.check_authentication(endpoint)
        self.check_content_length()
        self.check_allowed_methods(endpoint)
        self.check_headers(endpoint)
        return endpoint()

    @register_handler(r"^/taxii2/$", handles_own_auth=True)
    def discovery_handler(self):
        if context.account is None and not self.config["public_discovery"]:
            raise Unauthorized()
        response = {
            "title": self.config["title"],
        }
        for key in ["description", "contact"]:
            if self.config.get(key):
                response[key] = self.config.get(key)
        default_api_root, api_roots = self.persistence.get_api_roots()
        if default_api_root:
            response["default"] = f"/taxii2/{default_api_root.id}/"
        response["api_roots"] = [f"/taxii2/{api_root.id}/" for api_root in api_roots]
        return make_taxii2_response(response)

    @register_handler(r"^/taxii2/(?P<api_root_id>[^/]+)/$", handles_own_auth=True)
    def api_root_handler(self, api_root_id):
        try:
            api_root = self.persistence.get_api_root(api_root_id=api_root_id)
        except DoesNotExistError:
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if context.account is None and not api_root.is_public:
            raise Unauthorized()
        response = {
            "title": api_root.title,
            "versions": ["application/taxii+json;version=2.1"],
            "max_content_length": self.config["max_content_length"],
        }
        if api_root.description:
            response["description"] = api_root.description
        return make_taxii2_response(response)

    @register_handler(r"^/taxii2/(?P<api_root_id>[^/]+)/status/(?P<job_id>[^/]+)/$")
    def job_handler(self, api_root_id, job_id):
        try:
            job, job_details = self.persistence.get_job_and_details(
                api_root_id=api_root_id, job_id=job_id
            )
        except DoesNotExistError:
            raise NotFound()
        response = {
            "id": job.id,
            "status": job.status,
            "request_timestamp": taxii2_datetimeformat(job.request_timestamp),
            "total_count": job_details.total_count,
            "success_count": len(job_details.success),
            "successes": [
                job_detail.as_taxii2_dict() for job_detail in job_details.success
            ],
            "failure_count": len(job_details.failure),
            "failures": [
                job_detail.as_taxii2_dict() for job_detail in job_details.failure
            ],
            "pending_count": len(job_details.pending),
            "pendings": [
                job_detail.as_taxii2_dict() for job_detail in job_details.pending
            ],
        }
        return make_taxii2_response(response)

    @register_handler(r"^/taxii2/(?P<api_root_id>[^/]+)/collections/$", handles_own_auth=True)
    def collections_handler(self, api_root_id):
        try:
            api_root = self.persistence.get_api_root(api_root_id=api_root_id)
        except DoesNotExistError:
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if context.account is None and not api_root.is_public:
            raise Unauthorized()
        collections = self.persistence.get_collections(api_root_id=api_root_id)
        response = {}
        if collections:
            response["collections"] = []
            for collection in collections:
                data = {
                    "id": collection.id,
                    "title": collection.title,
                    "can_read": collection.can_read(context.account),
                    "can_write": collection.can_write(context.account),
                    "media_types": ["application/stix+json;version=2.1"],
                }
                for key in ["description", "alias"]:
                    value = getattr(collection, key, None)
                    if value:
                        data[key] = value
                response["collections"].append(data)
        return make_taxii2_response(response)

    @register_handler(
        r"^/taxii2/(?P<api_root_id>[^/]+)/collections/(?P<collection_id_or_alias>[^/]+)/$",
        handles_own_auth=True,
    )
    def collection_handler(self, api_root_id, collection_id_or_alias):
        try:
            collection = self.persistence.get_collection(
                api_root_id=api_root_id, collection_id_or_alias=collection_id_or_alias
            )
        except DoesNotExistError:
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if context.account is None and not collection.can_read(context.account):
            raise Unauthorized()
        response = {
            "id": collection.id,
            "title": collection.title,
            "can_read": collection.can_read(context.account),
            "can_write": collection.can_write(context.account),
            "media_types": ["application/stix+json;version=2.1"],
        }
        for key in ["description", "alias"]:
            value = getattr(collection, key, None)
            if value:
                response[key] = value
        return make_taxii2_response(response)

    @register_handler(
        r"^/taxii2/(?P<api_root_id>[^/]+)/collections/(?P<collection_id_or_alias>[^/]+)/manifest/$",
        handles_own_auth=True,
    )
    def manifest_handler(self, api_root_id, collection_id_or_alias):
        filter_params = validate_list_filter_params(request.args, self.persistence.api)
        try:
            manifest, more = self.persistence.get_manifest(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                **filter_params,
            )
        except (DoesNotExistError, NoReadPermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if manifest:
            response = {
                "more": more,
                "objects": [
                    {
                        "id": obj.id,
                        "date_added": taxii2_datetimeformat(obj.date_added),
                        "version": taxii2_datetimeformat(obj.version),
                        "media_type": f"application/stix+json;version={obj.spec_version}",
                    }
                    for obj in manifest
                ],
            }
            headers = {
                "X-TAXII-Date-Added-First": min(
                    obj["date_added"] for obj in response["objects"]
                ),
                "X-TAXII-Date-Added-Last": max(
                    obj["date_added"] for obj in response["objects"]
                ),
            }
        else:
            response = {}
            headers = {}
        return make_taxii2_response(
            response,
            extra_headers=headers,
        )

    @register_handler(
        r"^/taxii2/(?P<api_root_id>[^/]+)/collections/(?P<collection_id_or_alias>[^/]+)/objects/$",
        ("GET", "POST"),
        valid_content_types=("application/taxii+json;version=2.1",),
        handles_own_auth=True,
    )
    def objects_handler(self, api_root_id, collection_id_or_alias):
        if request.method == "GET":
            return self.objects_get_handler(api_root_id, collection_id_or_alias)
        if request.method == "POST":
            return self.objects_post_handler(api_root_id, collection_id_or_alias)

    def objects_get_handler(self, api_root_id, collection_id_or_alias):
        filter_params = validate_list_filter_params(request.args, self.persistence.api)
        try:
            objects, more, next_param = self.persistence.get_objects(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                **filter_params,
            )
        except (DoesNotExistError, NoReadPermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if objects:
            response = {
                "more": more,
                "objects": [
                    {
                        "id": obj.id,
                        "type": obj.type,
                        "spec_version": obj.type,
                        **obj.serialized_data,
                    }
                    for obj in objects
                ],
            }
            headers = {
                "X-TAXII-Date-Added-First": taxii2_datetimeformat(
                    min(obj.date_added for obj in objects)
                ),
                "X-TAXII-Date-Added-Last": taxii2_datetimeformat(
                    max(obj.date_added for obj in objects)
                ),
            }
            if more:
                response["next"] = next_param
        else:
            response = {}
            headers = {}
        return make_taxii2_response(
            response,
            extra_headers=headers,
        )

    def objects_post_handler(self, api_root_id, collection_id_or_alias):
        validate_envelope(request.data)
        try:
            job, job_details = self.persistence.add_objects(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                data=request.get_json(),
            )
        except (DoesNotExistError, NoWritePermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        response = {
            "id": job.id,
            "status": job.status,
            "request_timestamp": taxii2_datetimeformat(job.request_timestamp),
            "total_count": job_details.total_count,
            "success_count": len(job_details.success),
            "successes": [
                job_detail.as_taxii2_dict() for job_detail in job_details.success
            ],
            "failure_count": len(job_details.failure),
            "failures": [
                job_detail.as_taxii2_dict() for job_detail in job_details.failure
            ],
            "pending_count": len(job_details.pending),
            "pendings": [
                job_detail.as_taxii2_dict() for job_detail in job_details.pending
            ],
        }
        headers = {}
        return make_taxii2_response(
            response,
            202,
            extra_headers=headers,
        )

    @register_handler(
        r"^/taxii2/(?P<api_root_id>[^/]+)/collections/(?P<collection_id_or_alias>[^/]+)/objects/(?P<object_id>[^/]+)/$",
        ("GET", "DELETE"),
        handles_own_auth=True,
    )
    def object_handler(self, api_root_id, collection_id_or_alias, object_id):
        if request.method == "GET":
            return self.object_get_handler(
                api_root_id, collection_id_or_alias, object_id
            )
        if request.method == "DELETE":
            return self.object_delete_handler(
                api_root_id, collection_id_or_alias, object_id
            )

    def object_get_handler(self, api_root_id, collection_id_or_alias, object_id):
        filter_params = validate_object_filter_params(
            request.args, self.persistence.api
        )
        try:
            versions, more, next_param = self.persistence.get_object(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                object_id=object_id,
                **filter_params,
            )
        except (DoesNotExistError, NoReadPermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if versions:
            response = {
                "more": more,
                "objects": [
                    {
                        "id": obj.id,
                        "type": obj.type,
                        "spec_version": obj.type,
                        **obj.serialized_data,
                    }
                    for obj in versions
                ],
            }
            headers = {
                "X-TAXII-Date-Added-First": taxii2_datetimeformat(
                    min(obj.date_added for obj in versions)
                ),
                "X-TAXII-Date-Added-Last": taxii2_datetimeformat(
                    max(obj.date_added for obj in versions)
                ),
            }
            if more:
                response["next"] = next_param
        else:
            response = {}
            headers = {}
        return make_taxii2_response(
            response,
            extra_headers=headers,
        )

    def object_delete_handler(self, api_root_id, collection_id_or_alias, object_id):
        filter_params = validate_delete_filter_params(request.args)
        try:
            self.persistence.delete_object(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                object_id=object_id,
                **filter_params,
            )
        except (DoesNotExistError, NoReadNoWritePermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        except (NoReadPermission, NoWritePermission):
            if context.account is None:
                raise Unauthorized()
            raise Forbidden()
        return make_taxii2_response("")

    @register_handler(
        (
            r"^/taxii2/(?P<api_root_id>[^/]+)/collections/(?P<collection_id_or_alias>[^/]+)"
            r"/objects/(?P<object_id>[^/]+)/versions/$"
        ),
        handles_own_auth=True,
    )
    def versions_handler(self, api_root_id, collection_id_or_alias, object_id):
        filter_params = validate_versions_filter_params(
            request.args, self.persistence.api
        )
        try:
            versions, more = self.persistence.get_versions(
                api_root_id=api_root_id,
                collection_id_or_alias=collection_id_or_alias,
                object_id=object_id,
                **filter_params,
            )
        except (DoesNotExistError, NoReadPermission):
            if context.account is None:
                raise Unauthorized()
            raise NotFound()
        if versions:
            response = {
                "more": more,
                "versions": [taxii2_datetimeformat(obj.version) for obj in versions],
            }
            headers = {
                "X-TAXII-Date-Added-First": taxii2_datetimeformat(
                    min(obj.date_added for obj in versions)
                ),
                "X-TAXII-Date-Added-Last": taxii2_datetimeformat(
                    max(obj.date_added for obj in versions)
                ),
            }
        else:
            response = {}
            headers = {}
        return make_taxii2_response(
            response,
            extra_headers=headers,
        )


class ServerMapping(NamedTuple):
    taxii1: Optional[TAXII1Server]
    taxii2: Optional[TAXII2Server]


class TAXIIServer:
    """
    TAXII Server class.

    This is the main entrypoint for http requests.
    It keeps the Auth API manager instance and dispatches requests and error handling to
    :class:`TAXII1Server`
    and
    :class:`TAXII2Server`
    """

    AUTH_MANAGER_CLASS: ClassVar[Type[AuthManager]] = AuthManager
    servers: ServerMapping
    app: Flask
    config: ServerConfig

    def __init__(self, config: ServerConfig):
        self.config = config
        servers_kwargs = {
            "taxii1": None,
            "taxii2": None,
        }
        if "taxii1" in config and config["taxii1"]:
            servers_kwargs["taxii1"] = TAXII1Server(
                {**config["taxii1"], "domain": config.get("domain")}
            )
        if "taxii2" in config and config["taxii2"]:
            servers_kwargs["taxii2"] = TAXII2Server(
                {**config["taxii2"], "domain": config.get("domain")}
            )
        self.servers = ServerMapping(**servers_kwargs)
        self.auth = self.AUTH_MANAGER_CLASS(
            server=self, api=initialize_api(config["auth_api"])
        )

    @property
    def real_servers(self):
        return [server for server in self.servers if server is not None]

    @property
    def current_server(self):
        try:
            server = context.taxiiserver
        except AttributeError:
            if len(self.real_servers) == 1:
                server = self.real_servers[0]
            else:
                server = None
        return server

    def init_app(self, app: Flask):
        """Connect taxii1, taxii2 and auth to flask."""
        self.app = app
        for server in self.real_servers:
            server.init_app(app)
        self.auth.api.init_app(app)

    def is_basic_auth_supported(self):
        """Check if basic auth is a supported feature."""
        return self.config.get("support_basic_auth", False)

    def get_endpoint(self, relative_path: str) -> Optional[Callable[[], Response]]:
        """Get first endpoint matching relative_path."""
        for server in self.real_servers:
            endpoint = server.get_endpoint(relative_path)
            if endpoint:
                endpoint.server = server
                return endpoint

    def handle_request(self, relative_path: str) -> Response:
        """Dispatch request to appropriate taxii* server."""
        relative_path = "/" + relative_path
        endpoint = self.get_endpoint(relative_path)
        if not endpoint:
            raise NotFound()
        context.taxiiserver = endpoint.server
        return endpoint()

    def handle_internal_error(self, error):
        """Dispatch internal error handling to appropriate taxii* server."""
        return context.taxiiserver.handle_internal_error(error)

    def handle_status_exception(self, error):
        """Dispatch status exception handling to appropriate taxii* server."""
        return context.taxiiserver.handle_status_exception(error)

    def handle_http_exception(self, error):
        """Dispatch http exception handling to appropriate taxii* server."""
        server = self.current_server
        if server:
            return server.handle_http_exception(error)
        else:
            return error.get_response()

    def handle_validation_exception(self, error):
        """Dispatch validation exception handling to appropriate taxii* server."""
        server = self.current_server
        if server:
            return server.handle_validation_exception(error)
        else:
            return error.get_response()

    def raise_unauthorized(self):
        """Dispatch unauthorized handling to appropriate taxii* server."""
        endpoint = self.get_endpoint(request.path)
        if endpoint:
            server = endpoint.server
        else:
            server = self.servers.taxii1
        context.taxiiserver = server
        return server.raise_unauthorized()



"""A module to put common database helper components."""
import uuid
from datetime import timezone

from sqlalchemy.dialects import mysql
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, DateTime, TypeDecorator


class GUID(TypeDecorator):
    """
    Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        """Switch implementation based on database dialect."""
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        """Convert from python to database representation."""
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        """Convert from database to python representation."""
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class UTCDateTime(TypeDecorator):
    """Platform-independent DateTime type that always stores and returns UTC."""

    impl = DateTime
    cache_ok = True

    def load_dialect_impl(self, dialect):
        """Switch implementation based on database dialect."""
        if dialect.name == 'mysql':
            return dialect.type_descriptor(mysql.DATETIME(fsp=6))
        else:
            return dialect.type_descriptor(DateTime())

    def process_bind_param(self, value, engine):
        """Convert from python to database representation."""
        if value is not None:
            return value.astimezone(timezone.utc)

    def process_result_value(self, value, engine):
        """Convert from database to python representation."""
        if value is not None:
            return value.replace(tzinfo=timezone.utc)



import six
import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10

from libtaxii.constants import (
    SVC_COLLECTION_MANAGEMENT, SVC_FEED_MANAGEMENT,
    VID_TAXII_SERVICES_10, VID_TAXII_SERVICES_11
)

from .entities import (
    ContentBindingEntity, InboxMessageEntity, ContentBlockEntity,
    ServiceEntity
)


def parse_content_binding(raw_content_binding, version):
    if version == 10:
        return ContentBindingEntity(
            binding=raw_content_binding,
            subtypes=None)
    if version == 11:
        return ContentBindingEntity(
            binding=raw_content_binding.binding_id,
            subtypes=raw_content_binding.subtype_ids)
    raise ValueError('invalid version')


def parse_content_bindings(bindings, version):
    return [parse_content_binding(b, version) for b in bindings]


def content_binding_entity_to_content_binding(content_binding, version):
    if version == 10:
        return content_binding.binding
    if version == 11:
        return tm11.ContentBinding(
            binding_id=content_binding.binding,
            subtype_ids=content_binding.subtypes)
    raise ValueError('invalid version')


def content_binding_entities_to_content_bindings(content_bindings, version):
    return [
        content_binding_entity_to_content_binding(c, version)
        for c in content_bindings]


def service_to_service_instances(service, version):
    service_instances = []

    for binding in service.supported_protocol_bindings:
        address = service.get_absolute_address(binding)

        if version == 10:

            stype = service.service_type
            if stype == SVC_COLLECTION_MANAGEMENT:
                stype = SVC_FEED_MANAGEMENT

            instance = tm10.ServiceInstance(
                service_type=stype,
                services_version=VID_TAXII_SERVICES_10,
                available=service.available,
                protocol_binding=binding,
                service_address=address,
                message_bindings=service.supported_message_bindings,
                message=service.description
            )
        elif version == 11:
            instance = tm11.ServiceInstance(
                service_type=service.service_type,
                services_version=VID_TAXII_SERVICES_11,
                available=service.available,
                protocol_binding=binding,
                service_address=address,
                message_bindings=service.supported_message_bindings,
                message=service.description
            )
        else:
            raise ValueError('invalid version')
        service_instances.append(instance)

    return service_instances


# PollingServiceInstance vs PollInstance
def poll_service_to_polling_service_instance(service, version,
                                             is_poll_instance_cls=False):

    instances = []

    module = tm11 if version == 11 else tm10

    if is_poll_instance_cls:
        cls = module.PollInstance
    else:
        cls = module.PollingServiceInstance

    for binding in service.supported_protocol_bindings:
        address = service.get_absolute_address(binding)
        instance = cls(
            poll_protocol=binding,
            poll_address=address,
            poll_message_bindings=service.supported_message_bindings)

        instances.append(instance)

    return instances


def subscription_service_to_subscription_method(service, version):

    instances = []

    module = tm11 if version == 11 else tm10

    for binding in service.supported_protocol_bindings:
        address = service.get_absolute_address(binding)
        instance = module.SubscriptionMethod(
            subscription_protocol=binding,
            subscription_address=address,
            subscription_message_bindings=service.supported_message_bindings
        )
        instances.append(instance)

    return instances


def inbox_to_receiving_inbox_instance(inbox):
    inbox_instances = []

    for protocol_binding in inbox.supported_protocol_bindings:

        inbox_instances.append(tm11.ReceivingInboxService(
            inbox_protocol=protocol_binding,
            inbox_address=inbox.get_absolute_address(protocol_binding),
            inbox_message_bindings=inbox.supported_message_bindings,
            supported_contents=inbox.get_supported_content(version=11)
        ))

    return inbox_instances


def collection_to_feedcollection_information(service, collection, version):

    polling_instances = []
    for poll in service.get_polling_services(collection):
        polling_instances.extend(
            poll_service_to_polling_service_instance(poll, version=version))

    push_methods = service.get_push_methods(collection)

    subscription_methods = []
    for s in service.get_subscription_services(collection):
        subscription_methods.extend(
            subscription_service_to_subscription_method(s, version=version))

    if collection.accept_all_content:
        supported_content = []
    else:
        supported_content = content_binding_entities_to_content_bindings(
            collection.supported_content, version=version)

    if version == 11:
        inbox_instances = []
        for inbox in service.get_receiving_inbox_services(collection):
            inbox_instances.extend(inbox_to_receiving_inbox_instance(inbox))

        return tm11.CollectionInformation(
            collection_name=collection.name,
            collection_description=collection.description,
            supported_contents=supported_content,
            available=collection.available,

            push_methods=push_methods,
            polling_service_instances=polling_instances,
            subscription_methods=subscription_methods,

            collection_volume=collection.volume,
            collection_type=collection.type,
            receiving_inbox_services=inbox_instances
        )

    if version == 10:
        return tm10.FeedInformation(
            feed_name=collection.name,
            feed_description=collection.description,
            supported_contents=supported_content,
            available=collection.available,

            push_methods=push_methods,
            polling_service_instances=polling_instances,
            subscription_methods=subscription_methods
            # collection_volume, collection_type, and
            # receiving_inbox_services are not supported in TAXII 1.0
        )

    raise ValueError('invalid version')


def subscription_to_subscription_instance(subscription, polling_services,
                                          version,
                                          subscription_parameters=None):

    polling_instances = []
    for poll in polling_services:
        polling_instances.extend(
            poll_service_to_polling_service_instance(
                poll, version=version, is_poll_instance_cls=True))

    params = dict(
        subscription_id=subscription.subscription_id,
        poll_instances=polling_instances,
    )

    if version == 10:
        params['delivery_parameters'] = None
        return tm10.SubscriptionInstance(**params)

    if version == 11:
        push_params = None

        params.update(dict(
            status=subscription.status,
            push_parameters=push_params,
        ))

        if subscription_parameters:
            bindings = content_binding_entities_to_content_bindings(
                subscription_parameters.content_bindings, version=version)

            params['subscription_parameters'] = tm11.SubscriptionParameters(
                response_type=subscription_parameters.response_type,
                content_bindings=bindings
            )

        return tm11.SubscriptionInstance(**params)

    raise ValueError('invalid version')


def inbox_message_to_inbox_message_entity(inbox_message, service_id, version):

    params = dict(
        message_id=inbox_message.message_id,

        # FIXME: how to get raw value?
        original_message=inbox_message.to_xml(),
        content_block_count=len(inbox_message.content_blocks),
        service_id=service_id
    )

    if version == 10:
        if inbox_message.subscription_information:
            si = inbox_message.subscription_information
            begin = si.inclusive_begin_timestamp_label
            end = si.inclusive_end_timestamp_label
            params.update(dict(
                subscription_collection_name=si.feed_name,
                subscription_id=si.subscription_id,

                # TODO: Match up exclusive vs inclusive
                exclusive_begin_timestamp_label=begin,
                inclusive_end_timestamp_label=end
            ))
        return InboxMessageEntity(**params)

    if version == 11:
        params.update(dict(
            result_id=inbox_message.result_id,
            destination_collections=inbox_message.destination_collection_names,
        ))
        if inbox_message.record_count:
            params.update(dict(
                record_count=inbox_message.record_count.record_count,
                partial_count=inbox_message.record_count.partial_count
            ))
        if inbox_message.subscription_information:
            si = inbox_message.subscription_information
            begin = si.exclusive_begin_timestamp_label
            end = si.inclusive_end_timestamp_label
            params.update(dict(
                subscription_collection_name=si.collection_name,
                subscription_id=si.subscription_id,
                exclusive_begin_timestamp_label=begin,
                inclusive_end_timestamp_label=end
            ))
        return InboxMessageEntity(**params)

    raise ValueError('invalid version')


def content_block_to_content_block_entity(content_block, version,
                                          inbox_message_id=None):

    content_binding = parse_content_binding(
        content_block.content_binding,
        version=version)

    message = content_block.message if version == 11 else None

    # TODO: What about signatures?
    return ContentBlockEntity(
        id=None,
        message=message,
        inbox_message_id=inbox_message_id,
        content=content_block.content,
        timestamp_label=content_block.timestamp_label,
        content_binding=content_binding
        # padding = content_block.padding,
    )


def content_block_entity_to_content_block(entity, version):
    content_bindings = content_binding_entity_to_content_binding(
        entity.content_binding,
        version=version)

    # Libtaxii requires content to be unicode
    content = (
        entity.content
        if isinstance(entity.content, six.string_types)
        else entity.content.decode('utf-8'))

    if version == 10:
        return tm10.ContentBlock(
            content_binding=content_bindings,
            content=content,
            timestamp_label=entity.timestamp_label)
    if version == 11:
        return tm11.ContentBlock(
            content_binding=content_bindings,
            content=content,
            timestamp_label=entity.timestamp_label,
            message=entity.message)
    raise ValueError('invalid version')


def dict_to_service_entity(blob):
    properties = dict(blob)
    _id = properties.pop('id')
    _type = properties.pop('type')
    return ServiceEntity(id=_id, type=_type, properties=properties)



"""Taxii2 validation functions."""
import datetime
import json

from marshmallow import Schema, fields
from opentaxii.persistence.api import OpenTAXII2PersistenceAPI
from opentaxii.taxii2.exceptions import ValidationError
from opentaxii.taxii2.utils import DATETIMEFORMAT
from stix2 import parse
from stix2.exceptions import STIXError
from werkzeug.datastructures import ImmutableMultiDict


def validate_envelope(json_data: str, allow_custom: bool = False) -> None:
    """
    Validate if ``json_data`` is a valid taxii2 envelope.

    :param str json_data: the data to check
    :param bool allow_custom: if true, allow non-standard stix types
    """
    if not json_data:
        raise ValidationError("No data")
    try:
        data = json.loads(json_data)
    except json.JSONDecodeError as e:
        raise ValidationError(f"Invalid json: {str(e)}") from e
    if "objects" not in data:
        raise ValidationError("No objects")
    for item in data["objects"]:
        try:
            parse(item, allow_custom)
        except STIXError as e:
            raise ValidationError(
                f"Invalid stix object: {json.dumps(item)}; {str(e)}"
            ) from e


class Taxii2DateTime(fields.DateTime):
    """Taxii2 formatting compliant datetime field."""

    DEFAULT_FORMAT = DATETIMEFORMAT

    def _deserialize(self, value, attr, data, **kwargs):
        value = super()._deserialize(value, attr, data, **kwargs)
        return value.replace(tzinfo=datetime.timezone.utc)


class Taxii2Next(fields.Field):
    """Implemenatation of the taxii2 `next` query param."""

    def _deserialize(self, value, attr, data, **kwargs):
        value = super()._deserialize(value, attr, data, **kwargs)
        try:
            value = self.parent.persistence_api.parse_next_param(value)
        except:  # noqa
            raise ValidationError("Not a valid value.")
        return value


class Taxii2Filter(fields.Field):
    """General taxii2 filter implementation."""

    def _deserialize(self, value, attr, data, **kwargs):
        value = super()._deserialize(value, attr, data, **kwargs)
        return value.split(",")


class Taxii2VersionFilter(Taxii2Filter):
    """Taxii2 compliant version filter."""

    def _deserialize(self, value, attr, data, **kwargs):
        values = super()._deserialize(value, attr, data, **kwargs)
        new_values = []
        for value in values:
            if value not in ["first", "last", "all"]:
                try:
                    value = datetime.datetime.strptime(value, DATETIMEFORMAT).replace(
                        tzinfo=datetime.timezone.utc
                    )
                except ValueError:
                    pass
            new_values.append(value)
        return new_values


class PersistenceApiMxin:
    """Store persistence api on schema instance, to reference in `Taxii2Next`"""
    def __init__(self, persistence_api: OpenTAXII2PersistenceAPI, *args, **kwargs):
        self.persistence_api = persistence_api
        super().__init__(*args, **kwargs)


class VersionFilterParamsSchema(PersistenceApiMxin, Schema):
    """Schema for the versions endpoint filters."""

    limit = fields.Int()
    added_after = Taxii2DateTime()
    next_kwargs = Taxii2Next(data_key="next")
    match_spec_version = Taxii2Filter(data_key="match[spec_version]")


class ObjectFilterParamsSchema(PersistenceApiMxin, Schema):
    """Schema for the object endpoint filters."""

    limit = fields.Int()
    added_after = Taxii2DateTime()
    next_kwargs = Taxii2Next(data_key="next")
    match_spec_version = Taxii2Filter(data_key="match[spec_version]")
    match_version = Taxii2VersionFilter(data_key="match[version]")


class ListFilterParamsSchema(PersistenceApiMxin, Schema):
    """Schema for the object list endpoint filters."""

    limit = fields.Int()
    added_after = Taxii2DateTime()
    next_kwargs = Taxii2Next(data_key="next")
    match_spec_version = Taxii2Filter(data_key="match[spec_version]")
    match_version = Taxii2VersionFilter(data_key="match[version]")
    match_id = Taxii2Filter(data_key="match[id]")
    match_type = Taxii2Filter(data_key="match[type]")


class DeleteFilterParamsSchema(Schema):
    """Schema for the object delete endpoint filters."""

    match_version = Taxii2VersionFilter(data_key="match[version]")
    match_spec_version = Taxii2Filter(data_key="match[spec_version]")


def validate_object_filter_params(
    filter_params: ImmutableMultiDict, persistence_api: OpenTAXII2PersistenceAPI
) -> dict:
    """Validate and load filter params for the object endpoint."""
    parsed_params = ObjectFilterParamsSchema(persistence_api).load(filter_params)
    return parsed_params


def validate_list_filter_params(
    filter_params: ImmutableMultiDict, persistence_api: OpenTAXII2PersistenceAPI
) -> dict:
    """Validate and load filter params for the list endpoint."""
    parsed_params = ListFilterParamsSchema(persistence_api).load(filter_params)
    return parsed_params


def validate_versions_filter_params(
    filter_params: ImmutableMultiDict, persistence_api: OpenTAXII2PersistenceAPI
) -> dict:
    """Validate and load filter params for the versions endpoint."""
    parsed_params = VersionFilterParamsSchema(persistence_api).load(filter_params)
    return parsed_params


def validate_delete_filter_params(filter_params: ImmutableMultiDict) -> dict:
    """Validate and load filter params for the delete endpoint."""
    parsed_params = DeleteFilterParamsSchema().load(filter_params)
    return parsed_params



"""Taxii2 entities."""
from datetime import datetime
from typing import Optional

from opentaxii.common.entities import Entity
from opentaxii.entities import Account
from opentaxii.taxii2.utils import taxii2_datetimeformat


class ApiRoot(Entity):
    """
    TAXII2 API Root entity.

    :param str id: id of this API root
    :param bool default: indicator of default api root, should only be True once
    :param str title: human readable plain text name used to identify this API Root
    :param str description: human readable plain text description for this API Root
    :param bool is_public: whether this is a publicly readable API root
    """

    def __init__(
        self, id: str, default: bool, title: str, description: str, is_public: bool
    ):
        """Initialize ApiRoot."""
        self.id = id
        self.default = default
        self.title = title
        self.description = description
        self.is_public = is_public


class Collection(Entity):
    """
    TAXII2 Collection entity.

    :param str id: id of this collection
    :param str api_root_id: id of the :class:`ApiRoot` this collection belongs to
    :param str title: human readable plain text name used to identify this collection
    :param str description: human readable plain text description for this collection
    :param str alias: human readable collection name that can be used on systems to alias a collection id
    :param bool is_public: whether this is a publicly readable collection
    """

    def __init__(
        self,
        id: str,
        api_root_id: str,
        title: str,
        description: str,
        alias: str,
        is_public: bool,
    ):
        """Initialize Collection."""
        self.id = id
        self.api_root_id = api_root_id
        self.title = title
        self.description = description
        self.alias = alias
        self.is_public = is_public

    def can_read(self, account: Optional[Account]):
        """Determine if `account` is allowed to read from this collection."""
        return self.is_public or (
            account
            and (
                account.is_admin or "read" in set(account.permissions.get(self.id, []))
            )
        )

    def can_write(self, account: Optional[Account]):
        """Determine if `account` is allowed to write to this collection."""
        return account and (
            account.is_admin or "write" in set(account.permissions.get(self.id, []))
        )


class STIXObject(Entity):
    """
    TAXII2 STIXObject entity.

    :param str id: id of this stix object
    :param str collection_id: id of the :class:`Collection` this stix object belongs to
    :param str type: type of this stix object
    :param str spec_version: stix version this object matches
    :param datetime date_added: the date and time this object was added
    :param datetime version: the version of this object
    :param dict serialized_data: the payload of this object
    """

    def __init__(
        self,
        id: str,
        collection_id: str,
        type: str,
        spec_version: str,
        date_added: datetime,
        version: datetime,
        serialized_data: dict,
    ):
        """Initialize STIXObject."""
        self.id = id
        self.collection_id = collection_id
        self.type = type
        self.spec_version = spec_version
        self.date_added = date_added
        self.version = version
        self.serialized_data = serialized_data


class ManifestRecord(Entity):
    """
    TAXII2 ManifestRecord entity.

    This is a cut-down version of :class:`STIXObject`, for efficiency.

    :param str id: id of this stix object
    :param datetime date_added: the date and time this object was added
    :param datetime version: the version of this object
    :param str spec_version: stix version this object matches
    """

    def __init__(
        self,
        id: str,
        date_added: datetime,
        version: datetime,
        spec_version: str,
    ):
        """Initialize ManifestRecord."""
        self.id = id
        self.date_added = date_added
        self.version = version
        self.spec_version = spec_version


class VersionRecord(Entity):
    """
    TAXII2 VersionRecord entity.

    This is a cut-down version of :class:`STIXObject`, for efficiency.

    :param datetime date_added: the date and time this object was added
    :param datetime version: the version of this object
    """

    def __init__(
        self,
        date_added: datetime,
        version: datetime,
    ):
        """Initialize VersionRecord."""
        self.date_added = date_added
        self.version = version


class Job(Entity):
    """
    TAXII2 Job entity, called a "status resource" in taxii2 docs.

    :param str id: id of this job
    :param str api_root_id: id of the :class:`ApiRoot` this collection belongs to
    :param str status: status of this job
    :param datetime request_timestamp: the datetime of the request that this status resource is monitoring
    :param datetime completed_timestamp: the datetime of the completion of this job (used for cleanup)
    """

    def __init__(
        self,
        id: str,
        api_root_id: str,
        status: str,
        request_timestamp: datetime,
        completed_timestamp: datetime,
    ):
        """Initialize Job."""
        self.id = id
        self.api_root_id = api_root_id
        self.status = status
        self.request_timestamp = request_timestamp
        self.completed_timestamp = completed_timestamp


class JobDetail(Entity):
    """
    TAXII2 JobDetail entity, part of "status resource" in taxii2 docs.

    :param str id: id of this job detail
    :param str job_id: id of the job this detail belongs to
    :param str stix_id: id of the :class:`STIXObject` this detail tracks
    :param datetime version: the version of this object
    :param str message: message indicating more information about the object being created,
        its pending state, or why the object failed to be created.
    :param str status: status of this job
    """

    def __init__(
        self,
        id: str,
        job_id: str,
        stix_id: str,
        version: datetime,
        message: str,
        status: str,
    ):
        """Initialize JobDetail."""
        self.id = id
        self.job_id = job_id
        self.stix_id = stix_id
        self.version = version
        self.message = message
        self.status = status

    def as_taxii2_dict(self):
        """Turn this object into a taxii2 dict."""
        response = {"id": self.stix_id, "version": taxii2_datetimeformat(self.version)}
        if self.message:
            response["message"] = self.message
        return response




class ResultsNotReady(Exception):
    pass


class DoesNotExistError(Exception):
    pass


class NoReadPermission(Exception):
    pass


class NoWritePermission(Exception):
    pass


class NoReadNoWritePermission(NoReadPermission, NoWritePermission):
    pass



"""Taxii2 http helper functions."""
import json
from typing import Dict, Optional

from flask import Response, make_response


def make_taxii2_response(data, status: Optional[int] = 200, extra_headers: Optional[Dict] = None) -> Response:
    """Turn input data into valid taxii2 response."""
    if not isinstance(data, str):
        data = json.dumps(data)
    response = make_response((data, status))
    response.content_type = "application/taxii+json;version=2.1"
    response.headers.update(extra_headers or {})
    return response



import functools

import structlog
from flask import Flask, request
from marshmallow.exceptions import \
    ValidationError as MarshmallowValidationError
from werkzeug.exceptions import HTTPException

from .exceptions import InvalidAuthHeader
from .local import context, release_context
from .management import management
from .taxii.exceptions import StatusMessageException
from .taxii.http import HTTP_AUTHORIZATION
from .utils import parse_basic_auth_token

log = structlog.get_logger(__name__)


def create_app(server):
    """Create Flask application and attach TAXII server
    instance ``server`` to it.

    :param `opentaxii.server.TAXIIServer` server: TAXII server instance

    :return: Flask app
    """

    app = Flask(__name__)
    app.taxii_server = server

    server.init_app(app)

    app.add_url_rule(
        "/<path:relative_path>",
        "opentaxii_services_view",
        server.handle_request,
        methods=["GET", "POST", "OPTIONS", "DELETE"],
    )

    app.register_blueprint(management, url_prefix="/management")

    app.register_error_handler(500, server.handle_internal_error)
    app.register_error_handler(StatusMessageException, server.handle_status_exception)
    app.register_error_handler(HTTPException, server.handle_http_exception)
    app.register_error_handler(MarshmallowValidationError, server.handle_validation_exception)
    app.before_request(functools.partial(create_context_before_request, server))
    app.after_request(cleanup_context)
    return app


def create_context_before_request(server):
    context.account = _authenticate(server, request.headers)
    context.server = server


def cleanup_context(response):
    release_context()
    return response


def _authenticate(server, headers):

    auth_header = headers.get(HTTP_AUTHORIZATION)
    if not auth_header:
        return None

    parts = auth_header.split(" ", 1)

    if len(parts) != 2:
        log.warning("auth.header_invalid", value=auth_header)
        return None

    auth_type, raw_token = parts
    auth_type = auth_type.lower()

    if auth_type == "basic":

        if not server.is_basic_auth_supported():
            server.raise_unauthorized()

        try:
            username, password = parse_basic_auth_token(raw_token)
        except InvalidAuthHeader:
            log.error(
                "auth.basic_auth.header_invalid", raw_token=raw_token, exc_info=True
            )
            return None

        token = server.auth.authenticate(username, password)

    elif auth_type == "bearer":
        token = raw_token
    else:
        server.raise_unauthorized()

    if not token:
        server.raise_unauthorized()

    account = server.auth.get_account(token)

    if not account:
        server.raise_unauthorized()

    return account



import argparse

import structlog
import yaml
from opentaxii.cli import app
from opentaxii.entities import Account
from opentaxii.local import context
from opentaxii.utils import sync_conf_dict_into_db

log = structlog.getLogger(__name__)

local_admin = Account(id=None, username="local-admin", permissions=None, is_admin=True)


def sync_data_configuration():
    parser = argparse.ArgumentParser(
        description="Create services/collections/accounts",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("config", help="YAML file with data configuration")
    parser.add_argument(
        "-f",
        "--force-delete",
        dest="force_deletion",
        action="store_true",
        help=(
            "force deletion of collections and their content blocks "
            "if collection is not defined in configuration file"
        ),
        required=False,
    )
    args = parser.parse_args()
    with open(args.config) as stream:
        config = yaml.safe_load(stream=stream)

    with app.app_context():
        # run as admin with full access
        context.account = local_admin
        sync_conf_dict_into_db(
            app.taxii_server, config, force_collection_deletion=args.force_deletion
        )


def delete_content_blocks():

    parser = argparse.ArgumentParser(
        description=(
            "Delete content blocks from specified collections "
            "with timestamp labels matching defined time window"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--collection",
        action="append",
        dest="collection",
        help="Collection to remove content blocks from",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--with-messages",
        dest="delete_inbox_messages",
        action="store_true",
        help=("delete inbox messages associated with deleted content blocks"),
        required=False,
    )
    parser.add_argument(
        "--begin",
        dest="begin",
        help="exclusive beginning of time window as ISO8601 formatted date",
        required=True,
    )
    parser.add_argument(
        "--end",
        dest="end",
        help="inclusive ending of time window as ISO8601 formatted date",
    )

    args = parser.parse_args()
    with app.app_context():
        start_time = args.begin
        end_time = args.end
        for collection in args.collection:
            app.taxii_server.servers.taxii1.persistence.delete_content_blocks(
                collection,
                with_messages=args.delete_inbox_messages,
                start_time=start_time,
                end_time=end_time,
            )


def add_api_root():
    """CLI command to add taxii2 api root to database."""
    parser = argparse.ArgumentParser(
        description=("Add a new taxii2 ApiRoot object."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-t", "--title", required=True, help="Title of the api root")
    parser.add_argument(
        "-d", "--description", required=False, help="Description of the api root"
    )
    parser.add_argument(
        "--default", action="store_true", help="Set as default api root"
    )

    args = parser.parse_args()
    with app.app_context():
        app.taxii_server.servers.taxii2.persistence.api.add_api_root(
            title=args.title, description=args.description, default=args.default
        )


def add_collection():
    """CLI command to add taxii2 collection to database."""
    existing_api_root_ids = [
        str(api_root.id)
        for api_root in app.taxii_server.servers.taxii2.persistence.api.get_api_roots()
    ]
    parser = argparse.ArgumentParser(
        description=("Add a new taxii2 Collection object."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-r",
        "--rootid",
        choices=existing_api_root_ids,
        required=True,
        help="Api root id of the collection",
    )
    parser.add_argument("-t", "--title", required=True, help="Title of the collection")
    parser.add_argument(
        "-d", "--description", required=False, help="Description of the collection"
    )
    parser.add_argument("-a", "--alias", required=False, help="alias of the collection")
    parser.add_argument(
        "--public", action="store_true", help="allow public read access"
    )
    parser.set_defaults(public=False)

    args = parser.parse_args()
    with app.app_context():
        app.taxii_server.servers.taxii2.persistence.api.add_collection(
            api_root_id=args.rootid,
            title=args.title,
            description=args.description,
            alias=args.alias,
            is_public=args.public,
        )


def job_cleanup():
    """CLI command to clean up taxii2 job logs that are >24h old."""
    number_removed = app.taxii_server.servers.taxii2.persistence.api.job_cleanup()
    print(f"{number_removed} removed")



import datetime
from typing import Dict, List, Optional, Tuple

from opentaxii.taxii2.entities import (ApiRoot, Collection, Job, JobDetail,
                                       ManifestRecord, STIXObject,
                                       VersionRecord)


class OpenTAXIIPersistenceAPI:
    """Abstract class that represents OpenTAXII Persistence API.

    This class defines required methods that need to exist in
    a specific Persistence API implementation.
    """

    def init_app(self, app):
        pass

    def create_service(self, service_entity):
        """Create a service.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param `opentaxii.taxii.entities.ServiceEntity` service_entity:
            service entity in question
        :return: updated service entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.ServiceEntity`
        """
        raise NotImplementedError()

    def create_collection(self, collection_entity):
        """Create a collection.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param `opentaxii.taxii.entities.CollectionEntity` collection_entity:
            collection entity in question
        :return: updated collection entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        """
        raise NotImplementedError()

    def attach_collection_to_services(self, collection_id, service_ids):
        """Attach collection to the services.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param str collection_id: collection entity in question
        :param list service_ids: collection entity in question
        :return: updated collection entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        """
        raise NotImplementedError()

    def get_services(self, collection_id=None):
        """Get configured services.

        :param str collection_id: get only services assigned to
                                  collection with provided ID

        :return: list of service entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.ServiceEntity`
        """
        raise NotImplementedError()

    def get_collections(self, service_id=None):
        """Get the collections. If `service_id` is provided, return collection
        attached to a service.

        :param str service_id: ID of a service in question

        :return: list of collection entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.CollectionEntity`
        """
        raise NotImplementedError()

    def get_collection(self, collection_name, service_id=None):
        """Get a collection by name and service ID.

        Collection name is unique globally, so can be used as a key.
        Method retrieves collection entity using collection name
        ``collection_name`` and service ID ``service_id`` as a composite key.

        :param str collection_name: a collection name
        :param str service_id: ID of a service

        :return: collection entity
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        """
        raise NotImplementedError()

    def update_collection(self, collection_entity):
        """Update collection.

        :param `opentaxii.taxii.entities.CollectionEntity` collection_entity:
            collection entity object
        :return: updated collection entity
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        """
        raise NotImplementedError()

    def delete_collection(self, collection_name):
        """Delete collection.

        :param int collection_id: id of a collection object
        """
        pass

    def create_inbox_message(self, inbox_message_entity):
        """Create an inbox message.

        :param `opentaxii.taxii.entities.InboxMessageEntity` \
            inbox_message_entity: inbox message entity in question
        :return: updated inbox message entity
        :rtype: :py:class:`opentaxii.taxii.entities.InboxMessageEntity`
        """
        raise NotImplementedError()

    def create_content_block(
        self, content_block_entity, collection_ids=None, service_id=None
    ):
        """Create a content block.

        :param `opentaxii.taxii.entities.ContentBlockEntity` \
            content_block_entity: content block in question
        :param list collection_ids: a list of collection IDs as strings
        :param str service_id: ID of an inbox service via which content
            block was created

        :return: updated content block entity
        :rtype: :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        """
        raise NotImplementedError()

    def get_content_blocks_count(
        self, collection_id, start_time=None, end_time=None, bindings=None
    ):
        """Get a count of the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`

        :return: content block count
        :rtype: int
        """
        raise NotImplementedError()

    def get_content_blocks(
        self,
        collection_id,
        start_time=None,
        end_time=None,
        bindings=None,
        offset=0,
        limit=10,
    ):
        """Get the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`
        :param int offset: result set offset
        :param int limit: result set max size

        :return: content blocks list
        :rtype: list of :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        """
        raise NotImplementedError()

    def create_result_set(self, result_set_entity):
        """Create a result set.

        :param `opentaxii.taxii.entities.ResultSetEntity` result_set_entity:
            result set entity in question

        :return: updated result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        """
        raise NotImplementedError()

    def get_result_set(self, result_set_id):
        """Get a result set entity by ID.

        :param str result_set_id: ID of a result set.

        :return: result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        """
        raise NotImplementedError()

    def create_subscription(self, subscription_entity):
        """Create a subscription.

        :param `opentaxii.taxii.entities.SubscriptionEntity` \
            subscription_entity: subscription entity in question.

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        """
        raise NotImplementedError()

    def get_subscription(self, subscription_id):
        """Get a subscription entity by ID.

        :param str subscription_id: ID of a subscription

        :return: subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        """
        raise NotImplementedError()

    def get_subscriptions(self, service_id):
        """Get the subscriptions attached to/created via a service.

        :param str service_id: ID of a service

        :return: list of subscription entities
        :rtype: list of :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        """
        raise NotImplementedError()

    def update_subscription(self, subscription_entity):
        """Update a subscription status.

        :param `opentaxii.taxii.entities.SubscriptionEntity` \
            subscription_entity: subscription entity in question

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        """
        raise NotImplementedError()

    def get_domain(self, service_id):
        """Get configured domain needed to create absolute URLs.

        Returns `None` by default.

        :param str service_id: ID of a service
        """
        return None

    def delete_content_blocks(
        self, collection_name, start_time, end_time=None, with_messages=False
    ):
        """Delete content blocks in a specified collection with
        timestamp label in a specified time frame.

        :param str collection_name: collection name
        :param datetime start_time: exclusive beginning of a timeframe
        :param datetime end_time: inclusive end of a timeframe
        :param bool with_messages: delete related inbox messages
        """
        pass


class OpenTAXII2PersistenceAPI:
    """
    Abstract class that represents OpenTAXII Persistence API.

    Stub, pending implementation.
    """
    @staticmethod
    def get_next_param(self, kwargs: Dict) -> str:
        """
        Get value for `next` based on :class:`Dict` instance.

        :param :class:`Dict` kwargs: The dict to base the `next` param on

        :return: The value to use as `next` param
        :rtype: str
        """
        raise NotImplementedError

    @staticmethod
    def parse_next_param(next_param: str) -> Dict:
        """
        Parse provided `next_param` into kwargs to be used to filter stix objects.
        """
        raise NotImplementedError

    def get_api_roots(self) -> List[ApiRoot]:
        """
        Parse provided `next_param` into kwargs to be used to filter stix objects.
        """
        raise NotImplementedError

    def get_api_root(self, api_root_id: str) -> Optional[ApiRoot]:
        raise NotImplementedError

    def get_job_and_details(
        self, api_root_id: str, job_id: str
    ) -> Tuple[Optional[Job], List[JobDetail]]:
        raise NotImplementedError

    def get_collections(self, api_root_id: str) -> List[Collection]:
        raise NotImplementedError

    def get_collection(
        self, api_root_id: str, collection_id_or_alias: str
    ) -> Optional[Collection]:
        raise NotImplementedError

    def get_manifest(
        self,
        collection_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_id: Optional[List[str]] = None,
        match_type: Optional[List[str]] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[ManifestRecord], bool]:
        raise NotImplementedError

    def get_objects(
        self,
        collection_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_id: Optional[List[str]] = None,
        match_type: Optional[List[str]] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[STIXObject], bool, Optional[str]]:
        raise NotImplementedError

    def add_objects(self, api_root_id: str, collection_id: str, objects: List[Dict]) -> Tuple[Job, List[JobDetail]]:
        raise NotImplementedError

    def get_object(
        self,
        collection_id: str,
        object_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[Optional[List[STIXObject]], bool, Optional[str]]:
        """
        Get single object from database.

        Should return `None` when object matching object_id doesn't exist.
        """
        raise NotImplementedError

    def delete_object(
        self,
        collection_id: str,
        object_id: str,
        match_version: Optional[List[str]] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> None:
        raise NotImplementedError

    def get_versions(
        self,
        collection_id: str,
        object_id: str,
        limit: Optional[int] = None,
        added_after: Optional[datetime.datetime] = None,
        next_kwargs: Optional[Dict] = None,
        match_spec_version: Optional[List[str]] = None,
    ) -> Tuple[List[VersionRecord], bool]:
        """
        Get all versions of single object from database.

        Should return `None` when object matching object_id doesn't exist.
        """
        raise NotImplementedError



# flake8: noqa
from .api import OpenTAXII2PersistenceAPI, OpenTAXIIPersistenceAPI
from .manager import (BasePersistenceManager, Taxii1PersistenceManager,
                      Taxii2PersistenceManager)




class Account:
    '''Represents Account entity.

    This class holds user-specific information and is used
    for authorization.

    :param str id: account id
    :param dict details: additional details of an account
    '''

    def __init__(
            self, id, username, permissions, is_admin=False, **details):
        self.id = id
        self.username = username
        self.permissions = permissions
        self.is_admin = is_admin
        self.details = details

    def can_read(self, collection_name):
        return (
            self.is_admin or
            self.permissions.get(collection_name) in ('read', 'modify'))

    def can_modify(self, collection_name):
        return (
            self.is_admin or
            self.permissions.get(collection_name) == 'modify')

    def __repr__(self):
        return (
            'Account(username={}, is_admin={})'
            .format(self.username, self.is_admin))




class OpenTAXIIAuthAPI:
    '''Abstract class that represents OpenTAXII Authentication API.

    This class defines required methods that need to exist in
    a specific Authentication API implementation.
    '''

    def init_app(self, app):
        pass

    def authenticate(self, username, password):
        '''Authenticate a user.

        :param str username: username
        :param str password: password

        :return: auth token
        :rtype: string
        '''
        raise NotImplementedError()

    def get_account(self, token):
        '''Get account for auth token.

        :param str token: auth token

        :return: an account entity
        :rtype: `opentaxii.entities.Account`
        '''
        raise NotImplementedError()

    def create_account(self, account, password):
        '''Create an account.

        :param str username: username
        :param str password: password
        :param cool is_admin: is a new user admin?

        :return: an account entity
        :rtype: `opentaxii.entities.Account`
        '''
        raise NotImplementedError()

    def update_account(self, obj, password=None):
        '''Update an account.

        :param AccountEntity obj: an ipdated user entity (old one matched by username)
        :param str password: a new password

        :return: an account entity
        :rtype: `opentaxii.entities.Account`
        '''
        raise NotImplementedError()

    def get_accounts(self):
        raise NotImplementedError()

    def delete_account(self, username):
        raise NotImplementedError()



import structlog

import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10
from libtaxii.constants import (
    SD_SUPPORTED_CONTENT, ST_UNSUPPORTED_CONTENT_BINDING,
    SD_ITEM, ST_NOT_FOUND, ST_BAD_MESSAGE,
    ACT_SUBSCRIBE, ACT_UNSUBSCRIBE, ACT_PAUSE,
    ACT_RESUME, ACT_STATUS, ACT_TYPES_11,
    ACT_TYPES_10
)

from ...exceptions import StatusMessageException, raise_failure
from ...converters import (
    subscription_to_subscription_instance,
    parse_content_bindings
)
from ...entities import PollRequestParametersEntity, SubscriptionEntity

from .base_handlers import BaseMessageHandler
from .poll_request_handlers import retrieve_collection

log = structlog.getLogger(__name__)


def action_subscribe(request, service, collection, version, **kwargs):
    """
    Handle SUBSCRIBE action.

    taxii 1.0: create or update description
    taxii 1.1: first validate message, then create or update description
    """
    if version == 11:
        params = request.subscription_parameters
        response_type = params.response_type

        if not params.content_bindings:
            supported_contents = []
        else:
            requested_bindings = parse_content_bindings(
                params.content_bindings,
                version=version)

            supported_contents = \
                collection.get_matching_bindings(requested_bindings)

            if requested_bindings and not supported_contents:
                supported = collection.get_supported_content(version=version)
                details = {SD_SUPPORTED_CONTENT: supported}
                raise StatusMessageException(
                    ST_UNSUPPORTED_CONTENT_BINDING,
                    in_response_to=request.message_id,
                    status_details=details)

    else:
        supported_contents = []
        response_type = None

    poll_request_params = PollRequestParametersEntity(
        response_type=response_type,
        content_bindings=supported_contents,
    )

    # we are ignoring Delivery Parameters for now

    subscription = SubscriptionEntity(
        service_id=service.id,
        collection_id=collection.id,
        poll_request_params=poll_request_params,
        status=SubscriptionEntity.ACTIVE
    )

    return service.create_subscription(subscription)


def action_unsubscribe(request, service, subscription, **kwargs):
    """
    Handle UNSUBSCRIBE action.

    Update subscription to unsubscribed status, if it exists.
    """
    if subscription:
        subscription.status = SubscriptionEntity.UNSUBSCRIBED
        return service.update_subscription(subscription)
    # Spec says that unsubscribe should be successful even
    # if subscription doesn't exist
    return SubscriptionEntity(
        collection_id=None,
        service_id=service.id,
        subscription_id=request.subscription_id,
        status=SubscriptionEntity.UNSUBSCRIBED)


def action_status(service, subscription, **kwargs):
    """
    Handle STATUS action.

    Return status of single subscription or all subscriptions for service.
    """
    if subscription:
        return subscription
    return service.get_subscriptions()


def action_pause(service, subscription, **kwargs):
    """
    Handle PAUSE action.

    Update subscription to paused status, if it exists.
    """
    if subscription.status == SubscriptionEntity.PAUSED:
        return subscription
    subscription.status = SubscriptionEntity.PAUSED
    return service.update_subscription(subscription)


def action_resume(service, subscription, **kwargs):
    """
    Handle RESUME action.

    Update subscriptoin to active status, if it exists.
    """
    if subscription.status != SubscriptionEntity.PAUSED:
        return subscription
    subscription.status = SubscriptionEntity.ACTIVE
    return service.update_subscription(subscription)


ACTIONS = {
    ACT_SUBSCRIBE: action_subscribe,
    ACT_UNSUBSCRIBE: action_unsubscribe,
    ACT_PAUSE: action_pause,
    ACT_RESUME: action_resume,
    ACT_STATUS: action_status
}


class SubscriptionRequest11Handler(BaseMessageHandler):

    supported_request_messages = [tm11.ManageCollectionSubscriptionRequest]

    @classmethod
    def validate_request(cls, request, subscription):

        action = request.action

        if action not in ACT_TYPES_11:
            error_message = "The specified action was invalid"

        elif action in (ACT_UNSUBSCRIBE, ACT_PAUSE, ACT_RESUME) \
                and not request.subscription_id:
            error_message = 'Action "%s" requires a subscription id' % action

        else:
            error_message = None

        if error_message:
            raise StatusMessageException(
                ST_BAD_MESSAGE,
                message=error_message,
                in_response_to=request.message_id)

        if (not subscription and (
                action in (ACT_PAUSE, ACT_RESUME) or
                (action == ACT_STATUS and request.subscription_id))):

            details = {SD_ITEM: request.subscription_id}
            raise StatusMessageException(
                ST_NOT_FOUND,
                status_details=details,
                in_response_to=request.message_id)

    @classmethod
    def handle_message(cls, service, request):

        if request.subscription_id:
            subscription = service.get_subscription(request.subscription_id)
        else:
            subscription = None

        cls.validate_request(request, subscription)

        collection = retrieve_collection(
            11, service, request.collection_name, request.message_id)

        if subscription and subscription.collection_id != collection.id:
            raise StatusMessageException(
                ST_NOT_FOUND,
                status_details={SD_ITEM: request.collection_name},
                in_response_to=request.message_id)

        response = tm11.ManageCollectionSubscriptionResponse(
            message_id=cls.generate_id(),
            in_response_to=request.message_id,
            collection_name=collection.name,
            message=service.subscription_message,
        )

        result = ACTIONS[request.action](
            service=service, request=request, collection=collection,
            subscription=subscription, version=11)

        if isinstance(result, (list, tuple)):
            results = result
        else:
            results = [result]

        polling_services = service.get_polling_services(collection)

        for _result in results:
            instance = subscription_to_subscription_instance(
                subscription=_result,
                polling_services=polling_services,
                version=11,
                subscription_parameters=_result.params
            )

            response.subscription_instances.append(instance)

        return response


class SubscriptionRequest10Handler(BaseMessageHandler):
    supported_request_messages = [tm10.ManageFeedSubscriptionRequest]

    @classmethod
    def validate_request(cls, request):

        action = request.action

        if action not in ACT_TYPES_10:
            error_message = "The specified action was invalid"
        elif action == ACT_UNSUBSCRIBE and not request.subscription_id:
            error_message = 'Action "%s" requires a subscription id' % action
        else:
            error_message = None

        if error_message:
            raise StatusMessageException(
                ST_BAD_MESSAGE,
                message=error_message,
                in_response_to=request.message_id)

    @classmethod
    def handle_message(cls, service, request):

        cls.validate_request(request)

        collection = retrieve_collection(
            10, service, request.feed_name, request.message_id)

        if request.subscription_id:
            subscription = service.get_subscription(request.subscription_id)
        else:
            subscription = None

        if subscription and subscription.collection_id != collection.id:
            raise StatusMessageException(
                ST_NOT_FOUND,
                status_details=request.feed_name,
                in_response_to=request.message_id)

        response = tm10.ManageFeedSubscriptionResponse(
            message_id=cls.generate_id(),
            in_response_to=request.message_id,
            feed_name=collection.name,
            message=service.subscription_message,
        )

        results = ACTIONS[request.action](
            service=service,
            request=request,
            collection=collection,
            subscription=subscription,
            version=10)

        if not isinstance(results, (list, tuple)):
            results = [results]

        polling_services = service.get_polling_services(collection)

        for _result in results:
            instance = subscription_to_subscription_instance(
                subscription=_result,
                polling_services=polling_services,
                version=10)
            response.subscription_instances.append(instance)

        return response


class SubscriptionRequestHandler(BaseMessageHandler):

    supported_request_messages = [tm11.ManageCollectionSubscriptionRequest,
                                  tm10.ManageFeedSubscriptionRequest]

    @classmethod
    def handle_message(cls, service, request):
        if isinstance(request, tm10.ManageFeedSubscriptionRequest):
            return SubscriptionRequest10Handler.handle_message(
                service, request)
        if isinstance(request, tm11.ManageCollectionSubscriptionRequest):
            return SubscriptionRequest11Handler.handle_message(
                service, request)
        raise_failure(
            "TAXII Message not supported by message handler",
            request.message_id)



import structlog

log = structlog.getLogger(__name__)


class AuthManager:
    '''Manager responsible for authentication.

    Manager uses API instance ``api`` for basic auth operations and
    provides additional logic on top.

    :param `opentaxii.auth.api.OpenTAXIIAuthAPI` api:
        instance of Auth API class
    '''

    def __init__(self, server, api):
        self.server = server
        self.api = api

    def authenticate(self, username, password):
        '''Authenticate a user.

        :param str username: username
        :param str password: password

        :return: auth token
        :rtype: string
        '''
        return self.api.authenticate(username, password)

    def get_account(self, token):
        '''Get account for auth token.

        :param str token: auth token

        :return: an account entity
        :rtype: `opentaxii.entities.Account`
        '''
        return self.api.get_account(token)

    def update_account(self, account, password):
        '''Create an account.

        NOTE: Additional method that is only used in the helper scripts
        shipped with OpenTAXII.
        '''
        for colname, permission in list(account.permissions.items()):
            collection = self.server.servers.taxii1.persistence.get_collection(colname)
            if not collection:
                log.warning(
                    "update_account.unknown_collection",
                    collection=colname)
                account.permissions.pop(colname)
        account = self.api.update_account(account, password)
        return account

    def get_accounts(self):
        return self.api.get_accounts()

    def delete_account(self, username):
        return self.api.delete_account(username)



import os
from collections import defaultdict
from warnings import warn

import yaml
from libtaxii.constants import ST_TYPES_10, ST_TYPES_11

current_dir = os.path.dirname(os.path.realpath(__file__))

ENV_VAR_PREFIX = "OPENTAXII_"
CONFIG_ENV_VAR = "OPENTAXII_CONFIG"
DEFAULT_CONFIG_NAME = "defaults.yml"
DEFAULT_CONFIG = os.path.join(current_dir, DEFAULT_CONFIG_NAME)


def _infinite_dict():
    return defaultdict(_infinite_dict)


def merge(d1, d2):
    for k in d2:
        if k in d1 and isinstance(d1[k], dict) and isinstance(d2[k], dict):
            d1[k] = merge(d1[k], d2[k])
        else:
            d1[k] = d2[k]
    return d1


class ServerConfig(dict):
    """Class responsible for loading configuration files.

    This class will load default configuration file (shipped with OpenTAXII)
    and apply user specified configuration file on top of default one.

    Users can specify custom configuration file (YAML formatted) using
    enviromental variable. The variable should contain a full path to
    a custom configuration file.

    :param str optional_env_var: name of the enviromental variable
    :param list extra_configs: list of additional config filenames
    """

    VALID_BASE_OPTIONS = (
        "domain",
        "support_basic_auth",
        "return_server_error_details",
        "logging",
        "auth_api",
        "taxii1",
        "taxii2",
    )
    VALID_TAXII_OPTIONS = ("persistence_api",)
    VALID_TAXII1_OPTIONS = (
        "save_raw_inbox_messages",
        "xml_parser_supports_huge_tree",
        "count_blocks_in_poll_responses",
        "unauthorized_status",
        "hooks",
    )
    VALID_TAXII2_OPTIONS = (
        "contact",
        "description",
        "max_content_length",
        "title",
        "public_discovery",
    )
    ALL_VALID_OPTIONS = VALID_BASE_OPTIONS + VALID_TAXII_OPTIONS + VALID_TAXII1_OPTIONS

    def __init__(self, optional_env_var=CONFIG_ENV_VAR, extra_configs=None):

        # 4. default config
        configs = [DEFAULT_CONFIG]
        # 3. explicit configs
        configs.extend(extra_configs or [])
        # 2. config from OPENTAXII_CONFIG env var path
        env_var_path = os.environ.get(optional_env_var)
        if env_var_path:
            configs.append(env_var_path)
        # 1. config built from env vars
        configs.append(self._get_env_config(optional_env_var=optional_env_var))

        options = self._load_configs(*configs)
        options = self._clean_options(options)
        if (
            options["taxii1"]
            and options["taxii1"]["unauthorized_status"]
            not in ST_TYPES_10 + ST_TYPES_11
        ):
            raise ValueError("invalid value for unauthorized_status field")

        super(ServerConfig, self).__init__(options)

    @staticmethod
    def _get_env_config(env=os.environ, optional_env_var=None):
        result = _infinite_dict()
        for key, value in env.items():
            if not key.startswith(ENV_VAR_PREFIX):
                continue
            if key == optional_env_var:
                continue
            key = key[len(ENV_VAR_PREFIX):].lstrip("_").lower()
            value = yaml.safe_load(value)

            container = result
            parts = key.split("__")
            for part in parts[:-1]:
                container = container[part]
            container[parts[-1]] = value

        return dict(result)

    @classmethod
    def _load_configs(cls, *configs):
        result = dict()
        for config in configs:
            # read content from path-like object
            if not isinstance(config, dict):
                with open(config) as stream:
                    config = yaml.safe_load(stream=stream)
            result = merge(result, config)
        return result

    @classmethod
    def _clean_options(cls, options: dict) -> dict:
        # Put taxii and taxii2 keys that are in base level inside taxii1
        # structure for backwards compatibility
        for key in cls.VALID_TAXII1_OPTIONS + cls.VALID_TAXII_OPTIONS:
            if key in options:
                taxii1_dict = options.setdefault("taxii1", {})
                if taxii1_dict is None:
                    # taxii1 explicitly disabled -> taxii2-only mode
                    # drop taxii1 default values
                    warn(
                        f"Running in taxii2-only mode. Dropping deprecated top level taxii1 attribute '{key}'.",
                        DeprecationWarning,
                    )
                    del options[key]
                else:
                    warn(
                        f"Setting taxii1 attributes at top level is deprecated. Please nest '{key}' inside 'taxii1'.",
                        DeprecationWarning,
                    )
                    if key in taxii1_dict and isinstance(taxii1_dict[key], dict):
                        taxii1_dict[key] = merge(taxii1_dict[key], options.pop(key))
                    else:
                        taxii1_dict[key] = options.pop(key)
        # Warn user of invalid keys and remove from dict
        for key in [key for key in options if key not in cls.ALL_VALID_OPTIONS]:
            warn(f"Ignoring invalid configuration item '{key}'.")
            del options[key]
        if "taxii1" in options and options["taxii1"]:
            for key in [
                key
                for key in options["taxii1"]
                if key not in cls.VALID_TAXII_OPTIONS + cls.VALID_TAXII1_OPTIONS
            ]:
                warn(f"Ignoring invalid taxii1 configuration item '{key}'.")
                del options["taxii1"][key]
        if "taxii2" in options and options["taxii2"]:
            for key in [
                key
                for key in options["taxii2"]
                if key not in cls.VALID_TAXII_OPTIONS + cls.VALID_TAXII2_OPTIONS
            ]:
                warn(f"Ignoring invalid taxii2 configuration item '{key}'.")
                del options["taxii2"][key]
        return options
